"""Transport abstract base class definition"""
from abc import ABC, abstractmethod


class Transport(ABC):
    """Defines the operations that all transport classes should support"""
    @property
    @abstractmethod
    def connection_type(self):
        """The transport's connection type"""

    @property
    @abstractmethod
    def endpoint(self):
        """CometD service url"""

    @property
    @abstractmethod
    def client_id(self):
        """Clinet id value assigned by the server"""

    @property
    @abstractmethod
    def state(self):
        """Current state of the transport"""

    @property
    @abstractmethod
    def subscriptions(self):
        """Set of subscribed channels"""

    @property
    @abstractmethod
    def last_connect_result(self):
        """Result of the last connect request"""

    @abstractmethod
    async def handshake(self, connection_types):
        """Executes the handshake operation

        :param list[ConnectionType] connection_types: list of connection types
        :return: Handshake response
        :rtype: dict
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def connect(self):
        """Connect to the server

        The transport will try to start and maintain a continuous connection
        with the server, but it'll return with the response of the first
        successful connection as soon as possible.

        :return dict: The response of the first successful connection.
        :raise TransportInvalidOperation: If the transport doesn't has a \
        client id yet, or if it's not in a :obj:`~TransportState.DISCONNECTED`\
        :obj:`state`.
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def disconnect(self):
        """Disconnect from server

        The disconnect message is only sent to the server if the transport is
        actually connected.
        """

    @abstractmethod
    async def close(self):
        """Close transport and release resources"""

    @abstractmethod
    async def subscribe(self, channel):
        """Subscribe to *channel*

        :param str channel: Name of the channel
        :return: Subscribe response
        :rtype: dict
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def unsubscribe(self, channel):
        """Unsubscribe from *channel*

        :param str channel: Name of the channel
        :return: Unsubscribe response
        :rtype: dict
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def publish(self, channel, data):
        """Publish *data* to the given *channel*

        :param str channel: Name of the channel
        :param dict data: Data to send to the server
        :return: Publish response
        :rtype: dict
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def wait_for_state(self, state):
        """Waits for and returns when the transport enters the given *state*

        :param TransportState state: A state value
        """



import asyncio

from asynctest import TestCase, mock
from aiohttp import ClientSession

from aiocometd.transports.base import TransportBase
from aiocometd.constants import ConnectionType, MetaChannel, \
    TransportState, CONNECT_MESSAGE, SUBSCRIBE_MESSAGE, DISCONNECT_MESSAGE, \
    PUBLISH_MESSAGE, UNSUBSCRIBE_MESSAGE
from aiocometd.extensions import Extension, AuthExtension
from aiocometd.exceptions import TransportInvalidOperation, TransportError


class TransportBaseImpl(TransportBase):
    async def _send_final_payload(self, payload):
        pass

    @property
    def connection_type(self):
        return ConnectionType.LONG_POLLING


class TestTransportBase(TestCase):
    def setUp(self):
        self.transport = TransportBaseImpl(url="example.com/cometd",
                                           incoming_queue=None,
                                           loop=None)

    async def long_task(self, result, timeout=None):
        if timeout:
            await asyncio.sleep(timeout, loop=self.loop)
        if not isinstance(result, Exception):
            return result
        else:
            raise result

    def test_init_with_loop(self):
        loop = object()

        transport = TransportBaseImpl(url=None,
                                      incoming_queue=None,
                                      loop=loop)

        self.assertIs(transport._loop, loop)
        self.assertEqual(transport.state, TransportState.DISCONNECTED)

    @mock.patch("aiocometd.transports.base.asyncio")
    def test_init_without_loop(self, asyncio_mock):
        loop = object()
        asyncio_mock.get_event_loop.return_value = loop

        transport = TransportBaseImpl(url=None,
                                      incoming_queue=None)

        self.assertIs(transport._loop, loop)
        self.assertEqual(transport.state, TransportState.DISCONNECTED)

    async def test_get_http_session(self):
        self.transport._http_session = ClientSession()

        session = await self.transport._get_http_session()

        self.assertIsInstance(session, ClientSession)
        await session.close()

    async def test_get_http_session_creates_session(self):
        self.transport._http_session = None

        session = await self.transport._get_http_session()

        self.assertIsInstance(session, ClientSession)
        await session.close()

    @mock.patch("aiocometd.transports.base.asyncio")
    async def test_close_http_session(self, asyncio_mock):
        self.transport._http_session = mock.MagicMock()
        self.transport._http_session.closed = False
        self.transport._http_session.close = mock.CoroutineMock()
        asyncio_mock.sleep = mock.CoroutineMock()

        await self.transport._close_http_session()

        self.transport._http_session.close.assert_called()
        asyncio_mock.sleep.assert_called_with(
            self.transport._HTTP_SESSION_CLOSE_TIMEOUT)

    @mock.patch("aiocometd.transports.base.asyncio")
    async def test_close_http_session_already_closed(self, asyncio_mock):
        self.transport._http_session = mock.MagicMock()
        self.transport._http_session.closed = True
        self.transport._http_session.close = mock.CoroutineMock()
        asyncio_mock.sleep = mock.CoroutineMock()

        await self.transport._close_http_session()

        self.transport._http_session.close.assert_not_called()
        asyncio_mock.sleep.assert_not_called()

    def test_finalize_message_updates_fields(self):
        message = {
            "field": "value",
            "id": None,
            "clientId": None,
            "connectionType": None
        }
        self.transport._client_id = "client_id"

        self.transport._finalize_message(message)

        self.assertEqual(message["id"], str(0))
        self.assertEqual(self.transport._message_id, 1)
        self.assertEqual(message["clientId"], self.transport.client_id)
        self.assertEqual(message["connectionType"],
                         self.transport.connection_type.value)

    def test_finalize_message_ignores_non_existing_fields(self):
        message = {
            "field": "value"
        }
        self.transport._client_id = "client_id"

        self.transport._finalize_message(message)

        self.assertEqual(list(message.keys()), ["field"])
        self.assertEqual(message["field"], "value")

    def test_finalize_payload_single_message(self):
        payload = {
            "field": "value",
            "id": None,
            "clientId": None
        }
        self.transport._finalize_message = mock.MagicMock()

        self.transport._finalize_payload(payload)

        self.transport._finalize_message.assert_called_once_with(payload)

    def test_finalize_payload_multiple_messages(self):
        payload = [
            {
                "field": "value",
                "id": None,
                "clientId": None,
                "connectionType": None
            },
            {
                "field2": "value2",
                "id": None,
                "clientId": None,
                "connectionType": None
            }
        ]
        self.transport._finalize_message = mock.MagicMock()

        self.transport._finalize_payload(payload)

        self.transport._finalize_message.assert_has_calls([
            mock.call(payload[0]), mock.call(payload[1])
        ])

    @mock.patch("aiocometd.transports.base.is_event_message")
    @mock.patch("aiocometd.transports.base.is_server_error_message")
    async def test_consume_message(self, is_server_error_message,
                                   is_event_message):
        is_event_message.return_value = False
        is_server_error_message.return_value = False
        self.transport.incoming_queue = mock.MagicMock()
        self.transport.incoming_queue.put = mock.CoroutineMock()
        response_message = object()

        await self.transport._consume_message(response_message)

        is_server_error_message.assert_called_with(response_message)
        is_event_message.assert_called_with(response_message)
        self.transport.incoming_queue.put.assert_not_called()

    @mock.patch("aiocometd.transports.base.is_event_message")
    @mock.patch("aiocometd.transports.base.is_server_error_message")
    async def test_consume_message_event_message(self, is_server_error_message,
                                                 is_event_message):
        is_event_message.return_value = True
        is_server_error_message.return_value = False
        self.transport.incoming_queue = mock.MagicMock()
        self.transport.incoming_queue.put = mock.CoroutineMock()
        response_message = object()

        await self.transport._consume_message(response_message)

        is_server_error_message.assert_called_with(response_message)
        is_event_message.assert_called_with(response_message)

    @mock.patch("aiocometd.transports.base.is_event_message")
    @mock.patch("aiocometd.transports.base.is_server_error_message")
    async def test_consume_message_server_error_message(
            self, is_server_error_message, is_event_message):
        is_event_message.return_value = False
        is_server_error_message.return_value = True
        self.transport.incoming_queue = mock.MagicMock()
        self.transport.incoming_queue.put = mock.CoroutineMock()
        response_message = object()

        await self.transport._consume_message(response_message)

        is_server_error_message.assert_called_with(response_message)
        is_event_message.assert_not_called()

    def test_update_subscriptions_new_subscription_success(self):
        response_message = {
            "channel": MetaChannel.SUBSCRIBE,
            "subscription": "/test/channel1",
            "successful": True,
            "id": "3"
        }
        self.transport._subscriptions = set()

        self.transport._update_subscriptions(response_message)

        self.assertEqual(self.transport.subscriptions,
                         set([response_message["subscription"]]))

    def test_update_subscriptions_existing_subscription_success(self):
        response_message = {
            "channel": MetaChannel.SUBSCRIBE,
            "subscription": "/test/channel1",
            "successful": True,
            "id": "3"
        }
        self.transport._subscriptions = set([response_message["subscription"]])

        self.transport._update_subscriptions(response_message)

        self.assertEqual(self.transport.subscriptions,
                         set([response_message["subscription"]]))

    def test_update_subscriptions_new_subscription_fail(self):
        response_message = {
            "channel": MetaChannel.SUBSCRIBE,
            "subscription": "/test/channel1",
            "successful": False,
            "id": "3"
        }
        self.transport._subscriptions = set()

        self.transport._update_subscriptions(response_message)

        self.assertEqual(self.transport.subscriptions, set())

    def test_update_subscriptions_existing_subscription_fail(self):
        response_message = {
            "channel": MetaChannel.SUBSCRIBE,
            "subscription": "/test/channel1",
            "successful": False,
            "id": "3"
        }
        self.transport._subscriptions = set([response_message["subscription"]])

        self.transport._update_subscriptions(response_message)

        self.assertEqual(self.transport.subscriptions, set())

    def test_update_subscriptions_new_unsubscription_success(self):
        response_message = {
            "channel": MetaChannel.UNSUBSCRIBE,
            "subscription": "/test/channel1",
            "successful": True,
            "id": "3"
        }
        self.transport._subscriptions = set()

        self.transport._update_subscriptions(response_message)

        self.assertEqual(self.transport.subscriptions, set())

    def test_update_subscriptions_existing_unsubscription_success(self):
        response_message = {
            "channel": MetaChannel.UNSUBSCRIBE,
            "subscription": "/test/channel1",
            "successful": True,
            "id": "3"
        }
        self.transport._subscriptions = set([response_message["subscription"]])

        self.transport._update_subscriptions(response_message)

        self.assertEqual(self.transport.subscriptions, set())

    def test_update_subscriptions_new_unsubscription_fail(self):
        response_message = {
            "channel": MetaChannel.UNSUBSCRIBE,
            "subscription": "/test/channel1",
            "successful": False,
            "id": "3"
        }
        self.transport._subscriptions = set()

        self.transport._update_subscriptions(response_message)

        self.assertEqual(self.transport.subscriptions, set())

    def test_update_subscriptions_existing_unsubscription_fail(self):
        response_message = {
            "channel": MetaChannel.UNSUBSCRIBE,
            "subscription": "/test/channel1",
            "successful": False,
            "id": "3"
        }
        self.transport._subscriptions = set([response_message["subscription"]])

        self.transport._update_subscriptions(response_message)

        self.assertEqual(self.transport.subscriptions,
                         set([response_message["subscription"]]))

    @mock.patch("aiocometd.transports.base.is_matching_response")
    async def test_consume_payload_matching_without_advice(
            self, is_matching_response):
        payload = [
            {
                "channel": MetaChannel.CONNECT,
                "successful": True,
                "id": "1"
            }
        ]
        message = object()
        self.transport._update_subscriptions = mock.MagicMock()
        is_matching_response.return_value = True
        self.transport._consume_message = mock.MagicMock()

        result = await self.transport._consume_payload(
            payload, find_response_for=message)

        self.assertEqual(result, payload[0])
        self.assertEqual(self.transport._reconnect_advice, {})
        self.transport._update_subscriptions.assert_called_with(payload[0])
        is_matching_response.assert_called_with(payload[0], message)
        self.transport._consume_message.assert_not_called()

    async def test_process_incoming_payload(self):
        extension = mock.create_autospec(spec=Extension)
        auth = mock.create_autospec(spec=AuthExtension)
        self.transport._extensions = [extension]
        self.transport._auth = auth
        payload = object()
        headers = object()

        await self.transport._process_incoming_payload(payload, headers)

        extension.incoming.assert_called_with(payload, headers)
        auth.incoming.assert_called_with(payload, headers)

    @mock.patch("aiocometd.transports.base.is_matching_response")
    async def test_consume_payload_matching_without_advice_extension(
            self, is_matching_response):
        payload = [
            {
                "channel": MetaChannel.CONNECT,
                "successful": True,
                "id": "1"
            }
        ]
        message = object()
        self.transport._update_subscriptions = mock.MagicMock()
        is_matching_response.return_value = True
        self.transport._consume_message = mock.MagicMock()
        self.transport._process_incoming_payload = mock.CoroutineMock()
        headers = object()

        result = await self.transport._consume_payload(
            payload, headers=headers, find_response_for=message)

        self.assertEqual(result, payload[0])
        self.assertEqual(self.transport._reconnect_advice, {})
        self.transport._update_subscriptions.assert_called_with(payload[0])
        is_matching_response.assert_called_with(payload[0], message)
        self.transport._consume_message.assert_not_called()
        self.transport._process_incoming_payload.assert_called_with(payload,
                                                                    headers)

    @mock.patch("aiocometd.transports.base.is_matching_response")
    async def test_consume_payload_matching_with_advice(self,
                                                        is_matching_response):
        payload = [
            {
                "channel": MetaChannel.CONNECT,
                "successful": True,
                "advice": {"interval": 0, "reconnect": "retry"},
                "id": "1"
            }
        ]
        message = object()
        self.transport._update_subscriptions = mock.MagicMock()
        is_matching_response.return_value = True
        self.transport._consume_message = mock.MagicMock()
        self.transport._process_incoming_payload = mock.CoroutineMock()

        result = await self.transport._consume_payload(
            payload, find_response_for=message)

        self.assertEqual(result, payload[0])
        self.assertEqual(self.transport._reconnect_advice,
                         payload[0]["advice"])
        self.transport._update_subscriptions.assert_called_with(payload[0])
        is_matching_response.assert_called_with(payload[0], message)
        self.transport._consume_message.assert_not_called()
        self.transport._process_incoming_payload.assert_called_with(payload,
                                                                    None)

    @mock.patch("aiocometd.transports.base.is_matching_response")
    async def test_consume_payload_non_matching(self, is_matching_response):
        payload = [
            {
                "channel": MetaChannel.CONNECT,
                "successful": True,
                "id": "1"
            }
        ]
        message = None
        self.transport._update_subscriptions = mock.MagicMock()
        is_matching_response.return_value = False
        self.transport._consume_message = mock.CoroutineMock()
        self.transport._process_incoming_payload = mock.CoroutineMock()

        result = await self.transport._consume_payload(
            payload, find_response_for=message)

        self.assertIsNone(result)
        self.assertEqual(self.transport._reconnect_advice, {})
        self.transport._update_subscriptions.assert_called_with(payload[0])
        is_matching_response.assert_called_with(payload[0], message)
        self.transport._consume_message.assert_called_with(payload[0])
        self.transport._process_incoming_payload.assert_called_with(payload,
                                                                    None)

    async def test_send_message(self):
        message = {
            "channel": "/test/channel1",
            "data": {},
            "clientId": None,
            "id": "1"
        }
        response = object()
        self.transport._send_payload_with_auth = \
            mock.CoroutineMock(return_value=response)

        result = await self.transport._send_message(message,
                                                    field="value")

        self.assertIs(result, response)
        self.assertEqual(message["field"], "value")
        self.transport._send_payload_with_auth.assert_called_with([message])

    async def test_handshake(self):
        connection_types = [ConnectionType.WEBSOCKET]
        response = {
            "clientId": "id1",
            "successful": True
        }
        self.transport._send_message = \
            mock.CoroutineMock(return_value=response)
        message = {
            "channel": MetaChannel.HANDSHAKE,
            "version": "1.0",
            "supportedConnectionTypes": None,
            "minimumVersion": "1.0",
            "id": None
        }
        self.transport._subscribe_on_connect = False

        result = await self.transport.handshake(connection_types)

        self.assertEqual(result, response)
        final_connection_types = [ConnectionType.WEBSOCKET.value,
                                  self.transport.connection_type.value]
        self.transport._send_message.assert_called_with(
            message,
            supportedConnectionTypes=final_connection_types)
        self.assertEqual(self.transport.client_id, response["clientId"])
        self.assertTrue(self.transport._subscribe_on_connect)

    async def test_handshake_empty_connection_types(self):
        connection_types = []
        response = {
            "clientId": "id1",
            "successful": True
        }
        self.transport._send_message = \
            mock.CoroutineMock(return_value=response)
        message = {
            "channel": MetaChannel.HANDSHAKE,
            "version": "1.0",
            "supportedConnectionTypes": None,
            "minimumVersion": "1.0",
            "id": None
        }
        self.transport._subscribe_on_connect = False

        result = await self.transport.handshake(connection_types)

        self.assertEqual(result, response)
        final_connection_types = [self.transport.connection_type.value]
        self.transport._send_message.assert_called_with(
            message,
            supportedConnectionTypes=final_connection_types)
        self.assertEqual(self.transport.client_id, response["clientId"])
        self.assertTrue(self.transport._subscribe_on_connect)

    async def test_handshake_with_own_connection_type(self):
        connection_types = [self.transport.connection_type]
        response = {
            "clientId": "id1",
            "successful": True
        }
        self.transport._send_message = \
            mock.CoroutineMock(return_value=response)
        message = {
            "channel": MetaChannel.HANDSHAKE,
            "version": "1.0",
            "supportedConnectionTypes": None,
            "minimumVersion": "1.0",
            "id": None
        }
        self.transport._subscribe_on_connect = False

        result = await self.transport.handshake(connection_types)

        self.assertEqual(result, response)
        final_connection_types = [self.transport.connection_type.value]
        self.transport._send_message.assert_called_with(
            message,
            supportedConnectionTypes=final_connection_types)
        self.assertEqual(self.transport.client_id, response["clientId"])
        self.assertTrue(self.transport._subscribe_on_connect)

    async def test_handshake_failure(self):
        connection_types = [ConnectionType.WEBSOCKET]
        response = {
            "clientId": "id1",
            "successful": False
        }
        self.transport._send_message = \
            mock.CoroutineMock(return_value=response)
        message = {
            "channel": MetaChannel.HANDSHAKE,
            "version": "1.0",
            "supportedConnectionTypes": None,
            "minimumVersion": "1.0",
            "id": None
        }
        self.transport._subscribe_on_connect = False

        result = await self.transport.handshake(connection_types)

        self.assertEqual(result, response)
        final_connection_types = [ConnectionType.WEBSOCKET.value,
                                  self.transport.connection_type.value]
        self.transport._send_message.assert_called_with(
            message,
            supportedConnectionTypes=final_connection_types)
        self.assertEqual(self.transport.client_id, None)
        self.assertFalse(self.transport._subscribe_on_connect)

    def test_subscriptions(self):
        self.assertIs(self.transport.subscriptions,
                      self.transport._subscriptions)

    def test_subscriptions_read_only(self):
        with self.assertRaises(AttributeError):
            self.transport.subscriptions = {"channel1", "channel2"}

    async def test__connect(self):
        self.transport._subscribe_on_connect = False
        response = {
            "channel": MetaChannel.CONNECT,
            "successful": True,
            "advice": {"interval": 0, "reconnect": "retry"},
            "id": "2"
        }
        self.transport._send_payload_with_auth = \
            mock.CoroutineMock(return_value=response)

        result = await self.transport._connect()

        self.assertEqual(result, response)
        self.transport._send_payload_with_auth.assert_called_with(
            [CONNECT_MESSAGE])
        self.assertFalse(self.transport._subscribe_on_connect)

    async def test__connect_subscribe_on_connect(self):
        self.transport._subscribe_on_connect = True
        self.transport._subscriptions = {"/test/channel1", "/test/channel2"}
        response = {
            "channel": MetaChannel.CONNECT,
            "successful": True,
            "advice": {"interval": 0, "reconnect": "retry"},
            "id": "2"
        }
        additional_messages = []
        for subscription in self.transport.subscriptions:
            message = SUBSCRIBE_MESSAGE.copy()
            message["subscription"] = subscription
            additional_messages.append(message)
        self.transport._send_payload_with_auth = \
            mock.CoroutineMock(return_value=response)

        result = await self.transport._connect()

        self.assertEqual(result, response)
        self.transport._send_payload_with_auth.assert_called_with(
            [CONNECT_MESSAGE] + additional_messages)
        self.assertFalse(self.transport._subscribe_on_connect)

    async def test__connect_subscribe_on_connect_error(self):
        self.transport._subscribe_on_connect = True
        self.transport._subscriptions = {"/test/channel1", "/test/channel2"}
        response = {
            "channel": MetaChannel.CONNECT,
            "successful": False,
            "advice": {"interval": 0, "reconnect": "retry"},
            "id": "2"
        }
        additional_messages = []
        for subscription in self.transport.subscriptions:
            message = SUBSCRIBE_MESSAGE.copy()
            message["subscription"] = subscription
            additional_messages.append(message)
        self.transport._send_payload_with_auth = \
            mock.CoroutineMock(return_value=response)

        result = await self.transport._connect()

        self.assertEqual(result, response)
        self.transport._send_payload_with_auth.assert_called_with(
            [CONNECT_MESSAGE] + additional_messages)
        self.assertTrue(self.transport._subscribe_on_connect)

    def test_state(self):
        self.assertIs(self.transport.state,
                      self.transport._state)

    def test_state_read_only(self):
        with self.assertRaises(AttributeError):
            self.transport.state = TransportState.CONNECTING

    @mock.patch("aiocometd.transports.base.asyncio")
    async def test_start_connect_task(self, asyncio_mock):
        task = mock.MagicMock()
        asyncio_mock.ensure_future.return_value = task

        async def coro_func():
            pass
        coro = coro_func()

        result = self.transport._start_connect_task(coro)

        asyncio_mock.ensure_future.assert_called_with(
            coro,
            loop=self.transport._loop)
        self.assertEqual(self.transport._connect_task, task)
        task.add_done_callback.assert_called_with(self.transport._connect_done)
        self.assertEqual(result, task)
        await coro

    @mock.patch("aiocometd.transports.base.asyncio")
    async def test_stop_connect_task(self, asyncio_mock):
        self.transport._connect_task = mock.MagicMock()
        self.transport._connect_task.done.return_value = False
        asyncio_mock.wait = mock.CoroutineMock()

        await self.transport._stop_connect_task()

        self.transport._connect_task.cancel.assert_called()
        asyncio_mock.wait.assert_called_with([self.transport._connect_task])

    @mock.patch("aiocometd.transports.base.asyncio")
    async def test_stop_connect_task_with_none_task(self, asyncio_mock):
        self.transport._connect_task = None
        asyncio_mock.wait = mock.CoroutineMock()

        await self.transport._stop_connect_task()

        asyncio_mock.wait.assert_not_called()

    @mock.patch("aiocometd.transports.base.asyncio")
    async def test_stop_connect_task_with_done_task(self, asyncio_mock):
        self.transport._connect_task = mock.MagicMock()
        self.transport._connect_task.done.return_value = True
        asyncio_mock.wait = mock.CoroutineMock()

        await self.transport._stop_connect_task()

        asyncio_mock.wait.assert_not_called()

    async def test_connect_error_on_invalid_state(self):
        self.transport._client_id = "id"
        response = {
            "channel": MetaChannel.CONNECT,
            "successful": True,
            "advice": {"interval": 0, "reconnect": "retry"},
            "id": "2"
        }
        self.transport._connect = mock.CoroutineMock(return_value=response)
        for state in TransportState:
            if state not in [TransportState.DISCONNECTED,
                             TransportState.SERVER_DISCONNECTED]:
                self.transport._state = state
                with self.assertRaisesRegex(TransportInvalidOperation,
                                            "Can't connect to a server "
                                            "without disconnecting first."):
                    await self.transport.connect()

    async def test_connect_error_on_invalid_client_id(self):
        self.transport._client_id = ""

        with self.assertRaisesRegex(TransportInvalidOperation,
                                    "Can't connect to the server without a "
                                    "client id. Do a handshake first."):
            await self.transport.connect()

    async def test_connect(self):
        self.transport._client_id = "id"
        response = {
            "channel": MetaChannel.CONNECT,
            "successful": True,
            "advice": {"interval": 0, "reconnect": "retry"},
            "id": "2"
        }
        self.transport._connect = mock.CoroutineMock(return_value=response)
        self.transport._connect_done = mock.MagicMock()
        self.transport._state = TransportState.DISCONNECTED

        result = await self.transport.connect()

        self.assertEqual(result, response)
        self.assertEqual(self.transport.state, TransportState.CONNECTING)

    async def test_connect_in_server_disconnected_state(self):
        self.transport._client_id = "id"
        response = {
            "channel": MetaChannel.CONNECT,
            "successful": True,
            "advice": {"interval": 0, "reconnect": "retry"},
            "id": "2"
        }
        self.transport._connect = mock.CoroutineMock(return_value=response)
        self.transport._connect_done = mock.MagicMock()
        self.transport._state = TransportState.SERVER_DISCONNECTED

        result = await self.transport.connect()

        self.assertEqual(result, response)
        self.assertEqual(self.transport.state, TransportState.CONNECTING)

    async def test_connect_done_with_result(self):
        task = self.loop.create_future()
        task.set_result("result")
        self.transport._follow_advice = mock.MagicMock()
        self.transport._state = TransportState.CONNECTING
        self.transport._reconnect_advice = {
            "interval": 1,
            "reconnect": "retry"
        }
        self.transport._reconnect_timeout = 2

        with self.assertLogs(TransportBase.__module__, "DEBUG") as log:
            self.transport._connect_done(task)

        log_message = "Connect task finished with: {!r}".format("result")
        self.assertEqual(
            log.output,
            ["DEBUG:{}:{}".format(TransportBase.__module__, log_message)])
        self.transport._follow_advice.assert_called_with(1)
        self.assertEqual(self.transport.state, TransportState.CONNECTED)

    async def test_connect_done_with_error(self):
        error = RuntimeError("error")
        task = self.loop.create_future()
        task.set_exception(error)
        self.transport._follow_advice = mock.MagicMock()
        self.transport._state = TransportState.CONNECTED
        self.transport._reconnect_advice = {
            "interval": 1,
            "reconnect": "retry"
        }
        self.transport._reconnect_timeout = 2

        with self.assertLogs(TransportBase.__module__, "DEBUG") as log:
            self.transport._connect_done(task)

        log_message = "Connect task finished with: {!r}".format(error)
        self.assertEqual(
            log.output,
            ["DEBUG:{}:{}".format(TransportBase.__module__, log_message)])
        self.transport._follow_advice.assert_called_with(2)
        self.assertEqual(self.transport.state, TransportState.CONNECTING)

    async def test_connect_dont_follow_advice_on_disconnecting(self):
        error = RuntimeError("error")
        task = self.loop.create_future()
        task.set_exception(error)
        self.transport._follow_advice = mock.MagicMock()
        self.transport._state = TransportState.DISCONNECTING
        self.transport._reconnect_advice = {
            "interval": 1,
            "reconnect": "retry"
        }
        self.transport._reconnect_timeout = 2

        with self.assertLogs(TransportBase.__module__, "DEBUG") as log:
            self.transport._connect_done(task)

        log_message = "Connect task finished with: {!r}".format(error)
        self.assertEqual(
            log.output,
            ["DEBUG:{}:{}".format(TransportBase.__module__, log_message)])
        self.transport._follow_advice.assert_not_called()

    @mock.patch("aiocometd.transports.base.defer")
    def test_follow_advice_handshake(self, defer):
        self.transport._reconnect_advice = {
            "interval": 1,
            "reconnect": "handshake"
        }
        self.transport.handshake = mock.MagicMock(return_value=object())
        self.transport._connect = mock.MagicMock(return_value=object())
        self.transport._start_connect_task = mock.MagicMock()
        defer.return_value = mock.MagicMock(return_value=object())

        self.transport._follow_advice(5)

        defer.assert_called_with(self.transport.handshake,
                                 delay=5,
                                 loop=self.transport._loop)
        defer.return_value.assert_called_with([self.transport.connection_type])
        self.transport._connect.assert_not_called()
        self.transport._start_connect_task.assert_called_with(
            defer.return_value.return_value
        )

    @mock.patch("aiocometd.transports.base.defer")
    def test_follow_advice_retry(self, defer):
        self.transport._reconnect_advice = {
            "interval": 1,
            "reconnect": "retry"
        }
        self.transport.handshake = mock.MagicMock(return_value=object())
        self.transport._connect = mock.MagicMock(return_value=object())
        self.transport._start_connect_task = mock.MagicMock()
        defer.return_value = mock.MagicMock(return_value=object())

        self.transport._follow_advice(5)

        defer.assert_called_with(self.transport._connect,
                                 delay=5,
                                 loop=self.transport._loop)
        defer.return_value.assert_called()
        self.transport.handshake.assert_not_called()
        self.transport._start_connect_task.assert_called_with(
            defer.return_value.return_value
        )

    @mock.patch("aiocometd.transports.base.defer")
    def test_follow_advice_none(self, defer):
        advices = ["none", "", None]
        for advice in advices:
            self.transport._state = TransportState.CONNECTED
            self.transport._reconnect_advice = {
                "interval": 1,
                "reconnect": advice
            }
            self.transport.handshake = mock.MagicMock(return_value=object())
            self.transport._connect = mock.MagicMock(return_value=object())
            self.transport._start_connect_task = mock.MagicMock()
            defer.return_value = mock.MagicMock()

            with self.assertLogs(TransportBase.__module__, "DEBUG") as log:
                self.transport._follow_advice(5)

            self.assertEqual(log.output,
                             ["WARNING:{}:No reconnect "
                              "advice provided, no more operations will be "
                              "scheduled.".format(TransportBase.__module__)])
            defer.assert_not_called()
            self.transport.handshake.assert_not_called()
            self.transport._connect.assert_not_called()
            self.transport._start_connect_task.assert_not_called()
            self.assertEqual(self.transport.state,
                             TransportState.SERVER_DISCONNECTED)

    def test_client_id(self):
        self.assertIs(self.transport.client_id,
                      self.transport._client_id)

    def test_client_id_read_only(self):
        with self.assertRaises(AttributeError):
            self.transport.client_id = "id"

    def test_endpoint(self):
        self.assertIs(self.transport.endpoint,
                      self.transport._url)

    def test_endpoint_read_only(self):
        with self.assertRaises(AttributeError):
            self.transport.endpoint = ""

    async def test_disconnect(self):
        self.transport._state = TransportState.CONNECTED
        self.transport._stop_connect_task = mock.CoroutineMock()
        self.transport._send_message = mock.CoroutineMock()

        await self.transport.disconnect()

        self.assertEqual(self.transport.state, TransportState.DISCONNECTED)
        self.transport._stop_connect_task.assert_called()
        self.transport._send_message.assert_called_with(
            DISCONNECT_MESSAGE)

    async def test_disconnect_transport_error(self):
        self.transport._state = TransportState.CONNECTED
        self.transport._stop_connect_task = mock.CoroutineMock()
        self.transport._send_message = mock.CoroutineMock(
            side_effect=TransportError())

        await self.transport.disconnect()

        self.assertEqual(self.transport.state, TransportState.DISCONNECTED)
        self.transport._stop_connect_task.assert_called()
        self.transport._send_message.assert_called_with(
            DISCONNECT_MESSAGE)

    async def test_disconnect_if_not_connected(self):
        for state in (s for s in TransportState
                      if s != TransportState.CONNECTED):
            self.transport._state = state
            self.transport._stop_connect_task = mock.CoroutineMock()
            self.transport._send_message = mock.CoroutineMock()

            await self.transport.disconnect()

            self.assertEqual(self.transport.state, TransportState.DISCONNECTED)
            self.transport._stop_connect_task.assert_called()
            self.transport._send_message.assert_not_called()

    async def test_subscribe(self):
        for state in [TransportState.CONNECTED, TransportState.CONNECTING]:
            self.transport._state = state
            self.transport._send_message = mock.CoroutineMock(
                return_value="result"
            )

            result = await self.transport.subscribe("channel")

            self.assertEqual(result,
                             self.transport._send_message.return_value)
            self.transport._send_message.assert_called_with(
                SUBSCRIBE_MESSAGE,
                subscription="channel"
            )

    async def test_subscribe_error_if_not_connected(self):
        for state in TransportState:
            if state not in [TransportState.CONNECTED,
                             TransportState.CONNECTING]:
                self.transport._state = state
                self.transport._send_message = mock.CoroutineMock(
                    return_value="result"
                )

                with self.assertRaisesRegex(TransportInvalidOperation,
                                            "Can't subscribe without being "
                                            "connected to a server."):
                    await self.transport.subscribe("channel")

                self.transport._send_message.assert_not_called()

    async def test_unsubscribe(self):
        for state in [TransportState.CONNECTED, TransportState.CONNECTING]:
            self.transport._state = state
            self.transport._send_message = mock.CoroutineMock(
                return_value="result"
            )

            result = await self.transport.unsubscribe("channel")

            self.assertEqual(result,
                             self.transport._send_message.return_value)
            self.transport._send_message.assert_called_with(
                UNSUBSCRIBE_MESSAGE,
                subscription="channel"
            )

    async def test_unsubscribe_error_if_not_connected(self):
        for state in TransportState:
            if state not in [TransportState.CONNECTED,
                             TransportState.CONNECTING]:
                self.transport._state = state
                self.transport._send_message = mock.CoroutineMock(
                    return_value="result"
                )

                with self.assertRaisesRegex(TransportInvalidOperation,
                                            "Can't unsubscribe without being "
                                            "connected to a server."):
                    await self.transport.unsubscribe("channel")

                self.transport._send_message.assert_not_called()

    async def test_publish(self):
        for state in [TransportState.CONNECTED, TransportState.CONNECTING]:
            self.transport._state = state
            self.transport._send_message = mock.CoroutineMock(
                return_value="result"
            )

            result = await self.transport.publish("channel", {})

            self.assertEqual(result,
                             self.transport._send_message.return_value)
            self.transport._send_message.assert_called_with(
                PUBLISH_MESSAGE,
                channel="channel",
                data={}
            )

    async def test_publish_error_if_not_connected(self):
        for state in TransportState:
            if state not in [TransportState.CONNECTED,
                             TransportState.CONNECTING]:
                self.transport._state = state
                self.transport._send_message = mock.CoroutineMock(
                    return_value="result"
                )

                with self.assertRaisesRegex(TransportInvalidOperation,
                                            "Can't publish without being "
                                            "connected to a server."):
                    await self.transport.publish("channel", {})

                self.transport._send_message.assert_not_called()

    async def test_wait_for_state(self):
        state = TransportState.CONNECTING
        event = self.transport._state_events[state]
        event.wait = mock.CoroutineMock()

        await self.transport.wait_for_state(state)

        event.wait.assert_called()

    async def test_close(self):
        self.transport._close_http_session = mock.CoroutineMock()

        await self.transport.close()

        self.transport._close_http_session.assert_called()

    async def test_send_payload(self):
        payload = object()
        self.transport._finalize_payload = mock.MagicMock()
        response = object()
        self.transport._send_final_payload = mock.CoroutineMock(
            return_value=response
        )
        self.transport._process_outgoing_payload = mock.CoroutineMock()

        result = await self.transport._send_payload(payload)

        self.assertEqual(result, response)
        self.transport._finalize_payload.assert_called_with(payload)
        self.transport._send_final_payload.assert_called_with(payload,
                                                              headers={})
        self.transport._process_outgoing_payload.assert_called_with(payload,
                                                                    {})

    async def test_process_outgoing_payload(self):
        extension = mock.create_autospec(spec=Extension)
        auth = mock.create_autospec(spec=AuthExtension)
        self.transport._extensions = [extension]
        self.transport._auth = auth
        payload = object()
        headers = object()

        await self.transport._process_outgoing_payload(payload, headers)

        extension.outgoing.assert_called_with(payload, headers)
        auth.outgoing.assert_called_with(payload, headers)

    async def test_process_outgoing_payload_without_auth(self):
        extension = mock.create_autospec(spec=Extension)
        self.transport._extensions = [extension]
        self.transport._auth = None
        payload = object()
        headers = object()

        await self.transport._process_outgoing_payload(payload, headers)

        extension.outgoing.assert_called_with(payload, headers)

    async def test_send_payload_with_auth(self):
        response = object()
        payload = object()
        self.transport._send_payload = mock.CoroutineMock(
            return_value=response)
        self.transport._auth = None
        self.transport._is_auth_error_message = mock.MagicMock(
            return_value=False)

        result = await self.transport._send_payload_with_auth(payload)

        self.assertIs(result, response)
        self.transport._send_payload.assert_called_with(payload)
        self.transport._is_auth_error_message.assert_not_called()

    @mock.patch("aiocometd.transports.base.is_auth_error_message")
    async def test_send_payload_with_auth_with_extension(
            self, is_auth_error_message):
        response = object()
        payload = object()
        self.transport._send_payload = mock.CoroutineMock(
            return_value=response)
        self.transport._auth = mock.create_autospec(spec=AuthExtension)
        is_auth_error_message.return_value = False

        result = await self.transport._send_payload_with_auth(payload)

        self.assertIs(result, response)
        self.transport._send_payload.assert_called_with(payload)
        is_auth_error_message.assert_called_with(response)
        self.transport._auth.authenticate.assert_not_called()

    @mock.patch("aiocometd.transports.base.is_auth_error_message")
    async def test_send_payload_with_auth_with_extension_error(
            self, is_auth_error_message):
        response = object()
        response2 = object()
        payload = object()
        self.transport._send_payload = mock.CoroutineMock(
            side_effect=[response, response2])
        self.transport._auth = mock.create_autospec(spec=AuthExtension)
        is_auth_error_message.return_value = True

        result = await self.transport._send_payload_with_auth(payload)

        self.assertIs(result, response2)
        self.transport._send_payload.assert_has_calls([
            mock.call(payload), mock.call(payload)
        ])
        is_auth_error_message.assert_called_with(response)
        self.transport._auth.authenticate.assert_called()

    def test_state_sunder(self):
        result = self.transport._state

        self.assertEqual(result, self.transport.__dict__["_state"])

    def test_state_sunder_setter(self):
        state = TransportState.CONNECTED
        self.assertNotEqual(self.transport.__dict__["_state"], state)
        old_state = self.transport._state
        self.transport._set_state_event = mock.MagicMock()

        self.transport._state = state

        self.assertEqual(self.transport.__dict__["_state"], state)
        self.transport._set_state_event.assert_called_with(old_state, state)

    def test_set_state_event(self):
        old_state = TransportState.DISCONNECTED
        new_state = TransportState.CONNECTED
        self.transport._state_events[old_state].set()
        self.transport._state_events[new_state].clear()

        self.transport._set_state_event(old_state, new_state)

        self.assertFalse(self.transport._state_events[old_state].is_set())
        self.assertTrue(self.transport._state_events[new_state].is_set())

    def test_set_state_event_no_old_state(self):
        old_state = None
        new_state = TransportState.CONNECTED
        self.transport._state_events[new_state].clear()

        self.transport._set_state_event(old_state, new_state)

        self.assertTrue(self.transport._state_events[new_state].is_set())

    def test_set_state_event_unchanged_state(self):
        state = TransportState.CONNECTED
        event_mock = mock.MagicMock()
        self.transport._state_events[state] = event_mock

        self.transport._set_state_event(state, state)

        event_mock.set.assert_not_called()
        event_mock.clear.assert_not_called()

    def test_last_connect_result(self):
        self.transport._connect_task = mock.MagicMock()
        self.transport._connect_task.done.return_value = True
        self.transport._connect_task.result.return_value = object()

        result = self.transport.last_connect_result

        self.assertIs(result, self.transport._connect_task.result.return_value)

    def test_last_connect_result_on_no_connect_task(self):
        self.transport._connect_task = None

        result = self.transport.last_connect_result

        self.assertIsNone(result)



from unittest import TestCase, mock

from aiocometd.exceptions import ServerError


class TestServerError(TestCase):
    def test_properties(self):
        response = {
            "channel": "/meta/subscription",
            "successful": False,
            "id": "0",
            "error": "error"
        }

        error = ServerError("description message", response)

        self.assertEqual(error.message, "description message")
        self.assertEqual(error.response, response)
        self.assertEqual(error.error, "error")

    def test_properties_on_no_error(self):
        response = {
            "channel": "/meta/subscription",
            "successful": False,
            "id": "0"
        }

        error = ServerError("description message", response)

        self.assertEqual(error.message, "description message")
        self.assertEqual(error.response, response)
        self.assertEqual(error.error, None)

    @mock.patch("aiocometd.exceptions.utils")
    def test_error_code(self, utils):
        response = {
            "channel": "/meta/subscription",
            "successful": False,
            "id": "0",
            "error": "error"
        }
        utils.get_error_code.return_value = 12

        error = ServerError("description message", response)
        result = error.error_code

        self.assertEqual(result, utils.get_error_code.return_value)
        utils.get_error_code.assert_called_with(response["error"])

    @mock.patch("aiocometd.exceptions.utils")
    def test_error_message(self, utils):
        response = {
            "channel": "/meta/subscription",
            "successful": False,
            "id": "0",
            "error": "error"
        }
        utils.get_error_message.return_value = "message"

        error = ServerError("description message", response)
        result = error.error_message

        self.assertEqual(result, utils.get_error_message.return_value)
        utils.get_error_message.assert_called_with(response["error"])

    @mock.patch("aiocometd.exceptions.utils")
    def test_error_args(self, utils):
        response = {
            "channel": "/meta/subscription",
            "successful": False,
            "id": "0",
            "error": "error"
        }
        utils.get_error_args.return_value = ["arg"]

        error = ServerError("description message", response)
        result = error.error_args

        self.assertEqual(result, utils.get_error_args.return_value)
        utils.get_error_args.assert_called_with(response["error"])



import asyncio

from asynctest import TestCase, mock

from aiocometd.utils import get_error_message, get_error_code, get_error_args,\
    defer, is_auth_error_message, is_event_message, is_server_error_message, \
    is_matching_response
from aiocometd.constants import MetaChannel, SERVICE_CHANNEL_PREFIX


class TestGetErrorCode(TestCase):
    def test_get_error_code(self):
        error_field = "123::"

        result = get_error_code(error_field)

        self.assertEqual(result, 123)

    def test_get_error_code_none_field(self):
        error_field = None

        result = get_error_code(error_field)

        self.assertIsNone(result)

    def test_get_error_code_empty_field(self):
        error_field = ""

        result = get_error_code(error_field)

        self.assertIsNone(result)

    def test_get_error_code_invalid_field(self):
        error_field = "invalid"

        result = get_error_code(error_field)

        self.assertIsNone(result)

    def test_get_error_code_short_invalid_field(self):
        error_field = "12::"

        result = get_error_code(error_field)

        self.assertIsNone(result)

    def test_get_error_code_empty_code_field(self):
        error_field = "::"

        result = get_error_code(error_field)

        self.assertIsNone(result)


class TestGetErrorMessage(TestCase):
    def test_get_error_message(self):
        error_field = "::message"

        result = get_error_message(error_field)

        self.assertEqual(result, "message")

    def test_get_error_message_none_field(self):
        error_field = None

        result = get_error_message(error_field)

        self.assertIsNone(result)

    def test_get_error_message_empty_field(self):
        error_field = ""

        result = get_error_message(error_field)

        self.assertIsNone(result)

    def test_get_error_message_invalid_field(self):
        error_field = "invalid"

        result = get_error_message(error_field)

        self.assertIsNone(result)

    def test_get_error_message_empty_code_field(self):
        error_field = "::"

        result = get_error_message(error_field)

        self.assertEqual(result, "")


class TestGetErrorArgs(TestCase):
    def test_get_error_args(self):
        error_field = "403:xj3sjdsjdsjad,/foo/bar:Subscription denied"

        result = get_error_args(error_field)

        self.assertEqual(result, ["xj3sjdsjdsjad", "/foo/bar"])

    def test_get_error_args_none_field(self):
        error_field = None

        result = get_error_args(error_field)

        self.assertIsNone(result)

    def test_get_error_args_empty_field(self):
        error_field = ""

        result = get_error_args(error_field)

        self.assertIsNone(result)

    def test_get_error_args_invalid_field(self):
        error_field = "invalid"

        result = get_error_args(error_field)

        self.assertIsNone(result)

    def test_get_error_args_empty_code_field(self):
        error_field = "::"

        result = get_error_args(error_field)

        self.assertEqual(result, [])


class TestDefer(TestCase):
    def setUp(self):
        async def coro_func(value):
            return value

        self.coro_func = coro_func

    @mock.patch("aiocometd.utils.asyncio.sleep")
    async def test_defer(self, sleep):
        argument = object()
        delay = 10
        wrapper = defer(self.coro_func, delay, loop=self.loop)

        result = await wrapper(argument)

        self.assertIs(result, argument)
        sleep.assert_called_with(delay, loop=self.loop)

    @mock.patch("aiocometd.utils.asyncio.sleep")
    async def test_defer_no_loop(self, sleep):
        argument = object()
        delay = 10
        wrapper = defer(self.coro_func, delay)

        result = await wrapper(argument)

        self.assertIs(result, argument)
        sleep.assert_called_with(delay, loop=None)

    @mock.patch("aiocometd.utils.asyncio.sleep")
    async def test_defer_none_delay(self, sleep):
        argument = object()
        wrapper = defer(self.coro_func)

        result = await wrapper(argument)

        self.assertIs(result, argument)
        sleep.assert_not_called()

    @mock.patch("aiocometd.utils.asyncio.sleep")
    async def test_defer_zero_delay(self, sleep):
        argument = object()
        delay = 0
        wrapper = defer(self.coro_func, delay)

        result = await wrapper(argument)

        self.assertIs(result, argument)
        sleep.assert_not_called()

    @mock.patch("aiocometd.utils.asyncio.sleep")
    async def test_defer_sleep_canceled(self, sleep):
        argument = object()
        delay = 10
        wrapper = defer(self.coro_func, delay)
        sleep.side_effect = asyncio.CancelledError()

        with self.assertRaises(asyncio.CancelledError):
            await wrapper(argument)

        sleep.assert_called_with(delay, loop=None)


class TestIsAuthErrorMessage(TestCase):
    @mock.patch("aiocometd.utils.get_error_code")
    def test_is_auth_error_message(self, get_error_code):
        response = {
            "error": "error"
        }
        get_error_code.return_value = 401

        result = is_auth_error_message(response)

        self.assertTrue(result)
        get_error_code.assert_called_with(response["error"])

    @mock.patch("aiocometd.utils.get_error_code")
    def test_is_auth_error_message_forbidden(self, get_error_code):
        response = {
            "error": "error"
        }
        get_error_code.return_value = 403

        result = is_auth_error_message(response)

        self.assertTrue(result)
        get_error_code.assert_called_with(response["error"])

    @mock.patch("aiocometd.utils.get_error_code")
    def test_is_auth_error_message_not_an_auth_error(self, get_error_code):
        response = {
            "error": "error"
        }
        get_error_code.return_value = 400

        result = is_auth_error_message(response)

        self.assertFalse(result)
        get_error_code.assert_called_with(response["error"])

    @mock.patch("aiocometd.utils.get_error_code")
    def test_is_auth_error_message_not_an_error(self, get_error_code):
        response = {}
        get_error_code.return_value = None

        result = is_auth_error_message(response)

        self.assertFalse(result)
        get_error_code.assert_called_with(None)


class TestIsEventMessage(TestCase):
    def assert_event_message_for_channel(self, channel, has_data,
                                         expected_result):
        message = dict(channel=channel)
        if has_data:
            message["data"] = None

        result = is_event_message(message)

        self.assertEqual(result, expected_result)

    def test_is_event_message_subscribe(self):
        channel = MetaChannel.SUBSCRIBE
        self.assert_event_message_for_channel(channel, False, False)
        self.assert_event_message_for_channel(channel, True, False)

    def test_is_event_message_unsubscribe(self):
        channel = MetaChannel.UNSUBSCRIBE
        self.assert_event_message_for_channel(channel, False, False)
        self.assert_event_message_for_channel(channel, True, False)

    def test_is_event_message_non_meta_channel(self):
        channel = "/test/channel"
        self.assert_event_message_for_channel(channel, False, False)
        self.assert_event_message_for_channel(channel, True, True)

    def test_is_event_message_service_channel(self):
        channel = SERVICE_CHANNEL_PREFIX + "test"
        self.assert_event_message_for_channel(channel, False, False)
        self.assert_event_message_for_channel(channel, True, False)

    def test_is_event_message_handshake(self):
        channel = MetaChannel.HANDSHAKE
        self.assert_event_message_for_channel(channel, False, False)
        self.assert_event_message_for_channel(channel, True, False)

    def test_is_event_message_connect(self):
        channel = MetaChannel.CONNECT
        self.assert_event_message_for_channel(channel, False, False)
        self.assert_event_message_for_channel(channel, True, False)

    def test_is_event_message_disconnect(self):
        channel = MetaChannel.DISCONNECT
        self.assert_event_message_for_channel(channel, False, False)
        self.assert_event_message_for_channel(channel, True, False)


class TestIsServerErrorMessage(TestCase):
    def test_successful(self):
        message = {
            "successful": True
        }

        self.assertFalse(is_server_error_message(message))

    def test_not_successful(self):
        message = {
            "successful": False
        }

        self.assertTrue(is_server_error_message(message))

    def test_no_success_status(self):
        message = {}

        self.assertFalse(is_server_error_message(message))


class TestIsMatchingResponse(TestCase):
    def test_is_matching_response(self):
        message = {
            "channel": "/test/channel1",
            "data": {},
            "clientId": "clientId",
            "id": "1"
        }
        response = {
            "channel": "/test/channel1",
            "successful": True,
            "clientId": "clientId",
            "id": "1"
        }

        self.assertTrue(is_matching_response(response, message))

    def test_is_matching_response_response_none(self):
        message = {
            "channel": "/test/channel1",
            "data": {},
            "clientId": "clientId",
            "id": "1"
        }
        response = None

        self.assertFalse(is_matching_response(response, message))

    def test_is_matching_response_message_none(self):
        message = None
        response = {
            "channel": "/test/channel1",
            "successful": True,
            "clientId": "clientId",
            "id": "1"
        }

        self.assertFalse(is_matching_response(response, message))

    def test_is_matching_response_without_id(self):
        message = {
            "channel": "/test/channel1",
            "data": {},
            "clientId": "clientId",
        }
        response = {
            "channel": "/test/channel1",
            "successful": True,
            "clientId": "clientId",
        }

        self.assertTrue(is_matching_response(response, message))

    def test_is_matching_response_different_id(self):
        message = {
            "channel": "/test/channel1",
            "data": {},
            "clientId": "clientId",
            "id": "1"
        }
        response = {
            "channel": "/test/channel1",
            "successful": True,
            "clientId": "clientId",
            "id": "2"
        }

        self.assertFalse(is_matching_response(response, message))

    def test_is_matching_response_different_channel(self):
        message = {
            "channel": "/test/channel1",
            "data": {},
            "clientId": "clientId",
            "id": "1"
        }
        response = {
            "channel": "/test/channel2",
            "successful": True,
            "clientId": "clientId",
            "id": "1"
        }

        self.assertFalse(is_matching_response(response, message))

    def test_is_matching_response_without_successful_field(self):
        message = {
            "channel": "/test/channel1",
            "data": {},
            "clientId": "clientId",
            "id": "1"
        }
        response = {
            "channel": "/test/channel1",
            "clientId": "clientId",
            "id": "1"
        }

        self.assertFalse(is_matching_response(response, message))



"""Long polling transport class definition"""
import asyncio
import logging

import aiohttp

from ..constants import ConnectionType
from .registry import register_transport
from .base import TransportBase
from ..exceptions import TransportError


LOGGER = logging.getLogger(__name__)


@register_transport(ConnectionType.LONG_POLLING)
class LongPollingTransport(TransportBase):
    """Long-polling type transport"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        #: semaphore to limit the number of concurrent HTTP connections to 2
        self._http_semaphore = asyncio.Semaphore(2, loop=self._loop)

    async def _send_final_payload(self, payload, *, headers):
        try:
            session = await self._get_http_session()
            async with self._http_semaphore:
                response = await session.post(self._url, json=payload,
                                              ssl=self.ssl, headers=headers)
            response_payload = await response.json()
            headers = response.headers
        except aiohttp.client_exceptions.ClientError as error:
            LOGGER.warning("Failed to send payload, %s", error)
            raise TransportError(str(error)) from error
        return await self._consume_payload(response_payload, headers=headers,
                                           find_response_for=payload[0])



"""Package metadata"""
TITLE = "aiocometd"
DESCRIPTION = "CometD client for asyncio"
KEYWORDS = "asyncio aiohttp comet cometd bayeux push streaming"
URL = "https://github.com/robertmrk/aiocometd"
VERSION = "0.2.2"
AUTHOR = "Róbert Márki"
AUTHOR_EMAIL = "gsmiko@gmail.com"



from asynctest import TestCase, mock

from aiohttp import client_exceptions

from aiocometd.transports.long_polling import LongPollingTransport
from aiocometd.constants import ConnectionType
from aiocometd.exceptions import TransportError


class TestLongPollingTransport(TestCase):
    def setUp(self):
        self.transport = LongPollingTransport(url="example.com/cometd",
                                              incoming_queue=None,
                                              loop=None)

    def test_connection_type(self):
        self.assertEqual(self.transport.connection_type,
                         ConnectionType.LONG_POLLING)

    async def test_send_payload_final_payload(self):
        resp_data = [{
            "channel": "test/channel3",
            "data": {},
            "id": 4
        }]
        response_mock = mock.MagicMock()
        response_mock.json = mock.CoroutineMock(return_value=resp_data)
        response_mock.headers = object()
        session = mock.MagicMock()
        session.post = mock.CoroutineMock(return_value=response_mock)
        self.transport._get_http_session = \
            mock.CoroutineMock(return_value=session)
        self.transport._http_semaphore = mock.MagicMock()
        payload = [object(), object()]
        self.transport.ssl = object()
        self.transport._consume_payload = \
            mock.CoroutineMock(return_value=resp_data[0])
        headers = dict(key="value")

        response = await self.transport._send_final_payload(payload,
                                                            headers=headers)

        self.assertEqual(response, resp_data[0])
        self.transport._http_semaphore.__aenter__.assert_called()
        self.transport._http_semaphore.__aexit__.assert_called()
        session.post.assert_called_with(self.transport._url,
                                        json=payload,
                                        ssl=self.transport.ssl,
                                        headers=headers)
        self.transport._consume_payload.assert_called_with(
            resp_data,
            headers=response_mock.headers,
            find_response_for=payload[0])

    async def test_send_payload_final_payload_client_error(self):
        resp_data = [{
            "channel": "test/channel3",
            "data": {},
            "id": 4
        }]
        response_mock = mock.MagicMock()
        response_mock.json = mock.CoroutineMock(return_value=resp_data)
        session = mock.MagicMock()
        post_exception = client_exceptions.ClientError("client error")
        session.post = mock.CoroutineMock(side_effect=post_exception)
        self.transport._get_http_session = \
            mock.CoroutineMock(return_value=session)
        self.transport._http_semaphore = mock.MagicMock()
        payload = [object(), object()]
        self.transport.ssl = object()
        self.transport._consume_payload = \
            mock.CoroutineMock(return_value=resp_data[0])
        headers = dict(key="value")

        with self.assertLogs(LongPollingTransport.__module__,
                             level="DEBUG") as log:
            with self.assertRaisesRegex(TransportError, str(post_exception)):
                await self.transport._send_final_payload(payload,
                                                         headers=headers)

        log_message = "WARNING:{}:Failed to send payload, {}"\
            .format(LongPollingTransport.__module__, post_exception)
        self.assertEqual(log.output, [log_message])
        self.transport._http_semaphore.__aenter__.assert_called()
        self.transport._http_semaphore.__aexit__.assert_called()
        session.post.assert_called_with(self.transport._url,
                                        json=payload,
                                        ssl=self.transport.ssl,
                                        headers=headers)
        self.transport._consume_payload.assert_not_called()



from asynctest import TestCase, mock

from aiocometd.transports.registry import create_transport, \
    register_transport, TRANSPORT_CLASSES
from aiocometd.constants import ConnectionType
from aiocometd.exceptions import TransportInvalidOperation


class TestTransportFactoryFunctions(TestCase):
    def tearDown(self):
        TRANSPORT_CLASSES.clear()

    def test_register_transport(self):
        connection_type = ConnectionType.LONG_POLLING

        @register_transport(connection_type)
        class FakeTransport:
            "FakeTransport"

        obj = FakeTransport()

        self.assertEqual(obj.connection_type, connection_type)
        self.assertEqual(TRANSPORT_CLASSES[connection_type], FakeTransport)

    def test_create_transport(self):
        transport = object()
        transport_cls = mock.MagicMock(return_value=transport)
        TRANSPORT_CLASSES[ConnectionType.LONG_POLLING] = transport_cls

        result = create_transport(ConnectionType.LONG_POLLING,
                                  "arg", kwarg="value")

        self.assertEqual(result, transport)
        transport_cls.assert_called_with("arg", kwarg="value")

    def test_create_transport_error(self):
        connection_type = None

        with self.assertRaisesRegex(TransportInvalidOperation,
                                    "There is no transport for connection "
                                    "type {!r}".format(connection_type)):
            create_transport(connection_type)



"""Client class implementation"""
import asyncio
import reprlib
import logging
from collections import abc
from contextlib import suppress

from .transports import create_transport
from .constants import DEFAULT_CONNECTION_TYPE, ConnectionType, MetaChannel, \
    SERVICE_CHANNEL_PREFIX, TransportState
from .exceptions import ServerError, ClientInvalidOperation, \
    TransportTimeoutError, ClientError
from .utils import is_server_error_message


LOGGER = logging.getLogger(__name__)


class Client:  # pylint: disable=too-many-instance-attributes
    """CometD client"""
    #: Predefined server error messages by channel name
    _SERVER_ERROR_MESSAGES = {
        MetaChannel.HANDSHAKE: "Handshake request failed.",
        MetaChannel.CONNECT: "Connect request failed.",
        MetaChannel.DISCONNECT: "Disconnect request failed.",
        MetaChannel.SUBSCRIBE: "Subscribe request failed.",
        MetaChannel.UNSUBSCRIBE: "Unsubscribe request failed."
    }
    #: Defualt connection types list
    _DEFAULT_CONNECTION_TYPES = [ConnectionType.WEBSOCKET,
                                 ConnectionType.LONG_POLLING]

    def __init__(self, url, connection_types=None, *,
                 connection_timeout=10.0, ssl=None, max_pending_count=100,
                 extensions=None, auth=None, loop=None):
        """
        :param str url: CometD service url
        :param connection_types: List of connection types in order of \
        preference, or a single connection type name. If ``None``, \
        [:obj:`~ConnectionType.WEBSOCKET`, \
        :obj:`~ConnectionType.LONG_POLLING`] will be used as a default value.
        :type connection_types: list[ConnectionType], ConnectionType or None
        :param connection_timeout: The maximum amount of time to wait for the \
        transport to re-establish a connection with the server when the \
        connection fails.
        :type connection_timeout: int, float or None
        :param ssl: SSL validation mode. None for default SSL check \
        (:func:`ssl.create_default_context` is used), False for skip SSL \
        certificate validation, \
        `aiohttp.Fingerprint <https://aiohttp.readthedocs.io/en/stable/\
        client_reference.html#aiohttp.Fingerprint>`_ for fingerprint \
        validation, :obj:`ssl.SSLContext` for custom SSL certificate \
        validation.
        :param int max_pending_count: The maximum number of messages to \
        prefetch from the server. If the number of prefetched messages reach \
        this size then the connection will be suspended, until messages are \
        consumed. \
        If it is less than or equal to zero, the count is infinite.
        :param extensions: List of protocol extension objects
        :type extensions: list[Extension] or None
        :param AuthExtension auth: An auth extension
        :param loop: Event :obj:`loop <asyncio.BaseEventLoop>` used to
                     schedule tasks. If *loop* is ``None`` then
                     :func:`asyncio.get_event_loop` is used to get the default
                     event loop.
        """
        #: CometD service url
        self.url = url
        #: List of connection types to use in order of preference
        self._connection_types = None
        if isinstance(connection_types, ConnectionType):
            self._connection_types = [connection_types]
        elif isinstance(connection_types, abc.Iterable):
            self._connection_types = list(connection_types)
        else:
            self._connection_types = self._DEFAULT_CONNECTION_TYPES
        self._loop = loop or asyncio.get_event_loop()
        #: queue for consuming incoming event messages
        self._incoming_queue = None
        #: transport object
        self._transport = None
        #: marks whether the client is open or closed
        self._closed = True
        #: The maximum amount of time to wait for the transport to re-establish
        #: a connection with the server when the connection fails
        self.connection_timeout = connection_timeout
        #: SSL validation mode
        self.ssl = ssl
        #: the maximum number of messages to prefetch from the server
        self._max_pending_count = max_pending_count
        #: List of protocol extension objects
        self.extensions = extensions
        #: An auth extension
        self.auth = auth

    def __repr__(self):
        """Formal string representation"""
        cls_name = type(self).__name__
        fmt_spec = "{}({}, {}, connection_timeout={}, ssl={}, " \
                   "max_pending_count={}, extensions={}, auth={}, loop={})"
        return fmt_spec.format(cls_name,
                               reprlib.repr(self.url),
                               reprlib.repr(self._connection_types),
                               reprlib.repr(self.connection_timeout),
                               reprlib.repr(self.ssl),
                               reprlib.repr(self._max_pending_count),
                               reprlib.repr(self.extensions),
                               reprlib.repr(self.auth),
                               reprlib.repr(self._loop))

    @property
    def closed(self):
        """Marks whether the client is open or closed"""
        return self._closed

    @property
    def subscriptions(self):
        """Set of subscribed channels"""
        if self._transport:
            return self._transport.subscriptions
        return set()

    @property
    def connection_type(self):
        """The current connection type in use if the client is open,
        otherwise ``None``"""
        if self._transport is not None:
            return self._transport.connection_type
        return None

    @property
    def pending_count(self):
        """The number of pending incoming messages

        Once :obj:`open` is called the client starts listening for messages
        from the server. The incoming messages are retrieved and stored in an
        internal queue until they get consumed by calling :obj:`receive`.
        """
        if self._incoming_queue is None:
            return 0
        return self._incoming_queue.qsize()

    @property
    def has_pending_messages(self):
        """Marks whether the client has any pending incoming messages"""
        return self.pending_count > 0

    def _pick_connection_type(self, connection_types):
        """Pick a connection type based on the  *connection_types*
        supported by the server and on the user's preferences

        :param list[str] connection_types: Connection types \
        supported by the server
        :return: The connection type with the highest precedence \
        which is supported by the server
        :rtype: ConnectionType or None
        """
        server_connection_types = []
        for type_string in connection_types:
            with suppress(ValueError):
                server_connection_types.append(ConnectionType(type_string))

        intersection = (set(server_connection_types) &
                        set(self._connection_types))
        if not intersection:
            return None

        result = min(intersection, key=self._connection_types.index)
        return result

    async def _negotiate_transport(self):
        """Negotiate the transport type to use with the server and create the
        transport object

        :return: Transport object
        :rtype: Transport
        :raise ClientError: If none of the connection types offered by the \
        server are supported
        """
        self._incoming_queue = asyncio.Queue(maxsize=self._max_pending_count)
        transport = create_transport(DEFAULT_CONNECTION_TYPE,
                                     url=self.url,
                                     incoming_queue=self._incoming_queue,
                                     ssl=self.ssl,
                                     extensions=self.extensions,
                                     auth=self.auth,
                                     loop=self._loop)

        try:
            response = await transport.handshake(self._connection_types)
            self._verify_response(response)

            LOGGER.info("Connection types supported by the server: %r",
                        response["supportedConnectionTypes"])
            connection_type = self._pick_connection_type(
                response["supportedConnectionTypes"]
            )
            if not connection_type:
                raise ClientError("None of the connection types offered by "
                                  "the server are supported.")

            if transport.connection_type != connection_type:
                client_id = transport.client_id
                await transport.close()
                transport = create_transport(
                    connection_type,
                    url=self.url,
                    incoming_queue=self._incoming_queue,
                    client_id=client_id,
                    ssl=self.ssl,
                    extensions=self.extensions,
                    auth=self.auth,
                    loop=self._loop)
            return transport
        except Exception:
            await transport.close()
            raise

    async def open(self):
        """Establish a connection with the CometD server

        This method works mostly the same way as the `handshake` method of
        CometD clients in the reference implementations.

        :raise ClientError: If none of the connection types offered by the \
        server are supported
        :raise ClientInvalidOperation:  If the client is already open, or in \
        other words if it isn't :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the handshake or the first connect request \
        gets rejected by the server.
        """
        if not self.closed:
            raise ClientInvalidOperation("Client is already open.")

        LOGGER.info("Opening client with connection types %r ...",
                    [t.value for t in self._connection_types])
        self._transport = await self._negotiate_transport()

        response = await self._transport.connect()
        self._verify_response(response)
        self._closed = False
        LOGGER.info("Client opened with connection_type %r",
                    self.connection_type.value)

    async def close(self):
        """Disconnect from the CometD server"""
        if not self.closed:
            if self.pending_count == 0:
                LOGGER.info("Closing client...")
            else:
                LOGGER.warning(
                    "Closing client while %s messages are still pending...",
                    self.pending_count)
            try:
                if self._transport:
                    await self._transport.disconnect()
                    await self._transport.close()
            finally:
                self._closed = True
                LOGGER.info("Client closed.")

    async def subscribe(self, channel):
        """Subscribe to *channel*

        :param str channel: Name of the channel
        :raise ClientInvalidOperation: If the client is :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the subscribe request gets rejected by the \
        server
        """
        if self.closed:
            raise ClientInvalidOperation("Can't send subscribe request while, "
                                         "the client is closed.")
        await self._check_server_disconnected()

        response = await self._transport.subscribe(channel)
        self._verify_response(response)
        LOGGER.info("Subscribed to channel %s", channel)

    async def unsubscribe(self, channel):
        """Unsubscribe from *channel*

        :param str channel: Name of the channel
        :raise ClientInvalidOperation: If the client is :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the unsubscribe request gets rejected by the \
        server
        """
        if self.closed:
            raise ClientInvalidOperation("Can't send unsubscribe request "
                                         "while, the client is closed.")
        await self._check_server_disconnected()

        response = await self._transport.unsubscribe(channel)
        self._verify_response(response)
        LOGGER.info("Unsubscribed from channel %s", channel)

    async def publish(self, channel, data):
        """Publish *data* to the given *channel*

        :param str channel: Name of the channel
        :param dict data: Data to send to the server
        :return: Publish response
        :rtype: dict
        :raise ClientInvalidOperation: If the client is :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the publish request gets rejected by the server
        """
        if self.closed:
            raise ClientInvalidOperation("Can't publish data while, "
                                         "the client is closed.")
        await self._check_server_disconnected()

        response = await self._transport.publish(channel, data)
        self._verify_response(response)
        return response

    def _verify_response(self, response):
        """Check the ``successful`` status of the *response* and raise \
        the appropriate :obj:`~aiocometd.exceptions.ServerError` if it's False

        If the *response* has no ``successful`` field, it's considered to be
        successful.

        :param dict response: Response message
        :raise ServerError: If the *response* is not ``successful``
        """
        if is_server_error_message(response):
            self._raise_server_error(response)

    def _raise_server_error(self, response):
        """Raise the appropriate :obj:`~aiocometd.exceptions.ServerError` for \
        the failed *response*

        :param dict response: Response message
        :raise ServerError: If the *response* is not ``successful``
        """
        channel = response["channel"]
        message = type(self)._SERVER_ERROR_MESSAGES.get(channel)
        if not message:
            if channel.startswith(SERVICE_CHANNEL_PREFIX):
                message = "Service request failed."
            else:
                message = "Publish request failed."
        raise ServerError(message, response)

    async def receive(self):
        """Wait for incoming messages from the server

        :return: Incoming message
        :rtype: dict
        :raise ClientInvalidOperation: If the client is closed, and has no \
        more pending incoming messages
        :raise ServerError: If the client receives a confirmation message \
         which is not ``successful``
        :raise TransportTimeoutError: If the transport can't re-establish \
        connection with the server in :obj:`connection_timeout` time.
        """
        if not self.closed or self.has_pending_messages:
            response = await self._get_message(self.connection_timeout)
            self._verify_response(response)
            return response
        else:
            raise ClientInvalidOperation("The client is closed and there are "
                                         "no pending messages.")

    async def __aiter__(self):
        """Asynchronous iterator

        :raise ServerError: If the client receives a confirmation message \
         which is not ``successful``
        :raise TransportTimeoutError: If the transport can't re-establish \
        connection with the server in :obj:`connection_timeout` time.
        """
        while True:
            try:
                yield await self.receive()
            except ClientInvalidOperation:
                break

    async def __aenter__(self):
        """Enter the runtime context and call :obj:`open`

        :raise ClientInvalidOperation:  If the client is already open, or in \
        other words if it isn't :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the handshake or the first connect request \
        gets rejected by the server.
        :return: The client object itself
        :rtype: Client
        """
        try:
            await self.open()
        except Exception:
            await self.close()
            raise
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context and call :obj:`open`"""
        await self.close()

    async def _get_message(self, connection_timeout):
        """Get the next incoming message

        :param connection_timeout: The maximum amount of time to wait for the \
        transport to re-establish a connection with the server when the \
        connection fails.
        :return: Incoming message
        :rtype: dict
        :raise TransportTimeoutError: If the transport can't re-establish \
        connection with the server in :obj:`connection_timeout` time.
        :raise ServerError: If the connection gets closed by the server.
        """
        tasks = []
        # task waiting on connection timeout
        if connection_timeout:
            timeout_task = asyncio.ensure_future(
                self._wait_connection_timeout(connection_timeout),
                loop=self._loop
            )
            tasks.append(timeout_task)

        # task waiting on incoming messages
        get_task = asyncio.ensure_future(self._incoming_queue.get(),
                                         loop=self._loop)
        tasks.append(get_task)

        # task waiting on server side disconnect
        server_disconnected_task = asyncio.ensure_future(
            self._transport.wait_for_state(
                TransportState.SERVER_DISCONNECTED),
            loop=self._loop
        )
        tasks.append(server_disconnected_task)

        try:
            done, pending = await asyncio.wait(
                tasks,
                return_when=asyncio.FIRST_COMPLETED,
                loop=self._loop)

            # cancel all pending tasks
            for task in pending:
                task.cancel()

            # handle the completed task
            if get_task in done:
                return get_task.result()
            elif server_disconnected_task in done:
                await self._check_server_disconnected()
            else:
                raise TransportTimeoutError("Lost connection with the "
                                            "server.")
        except asyncio.CancelledError:
            # cancel all tasks
            for task in tasks:
                task.cancel()
            raise

    async def _wait_connection_timeout(self, timeout):
        """Wait for and return when the transport can't re-establish \
        connection with the server in *timeout* time

        :param timeout: The maximum amount of time to wait for the \
        transport to re-establish a connection with the server when the \
        connection fails.
        """
        while True:
            await self._transport.wait_for_state(TransportState.CONNECTING)
            try:
                await asyncio.wait_for(
                    self._transport.wait_for_state(TransportState.CONNECTED),
                    timeout, loop=self._loop
                )
            except asyncio.TimeoutError:
                break

    async def _check_server_disconnected(self):
        """Checks whether the current transport'state is
        :obj:`TransportState.SERVER_DISCONNECTED` and if it is then closes the
        client and raises an error

        :raise ServerError: If the current transport's state is \
        :obj:`TransportState.SERVER_DISCONNECTED`
        """
        if (self._transport and
                self._transport.state == TransportState.SERVER_DISCONNECTED):
            await self.close()
            raise ServerError("Connection closed by the server",
                              self._transport.last_connect_result)



from asynctest import TestCase, mock

from aiohttp import client_exceptions, WSMsgType

from aiocometd.transports.websocket import WebSocketTransport, WebSocketFactory
from aiocometd.constants import ConnectionType, MetaChannel
from aiocometd.exceptions import TransportConnectionClosed, TransportError


class TestWebSocketFactory(TestCase):
    def setUp(self):
        self.session = mock.CoroutineMock()
        self.session_factory = mock.CoroutineMock(return_value=self.session)
        self.url = "http://example.com/"
        self.factory = WebSocketFactory(self.session_factory)

    async def test_enter(self):
        socket = object()
        context = mock.MagicMock()
        context.__aenter__ = mock.CoroutineMock(return_value=socket)
        self.session.ws_connect.return_value = context
        args = [object()]
        kwargs = {"key": "value"}

        await self.factory._enter(*args, **kwargs)

        self.session.ws_connect.assert_called_with(*args, **kwargs)
        self.assertEqual(self.factory._context, context)
        self.assertEqual(self.factory._socket, socket)

    async def test_exit(self):
        socket = object()
        context = mock.MagicMock()
        context.__aexit__ = mock.CoroutineMock(return_value=socket)
        self.factory._context = context
        self.factory._socket = socket

        await self.factory._exit()

        context.__aexit__.assert_called()
        self.assertIsNone(self.factory._context)
        self.assertIsNone(self.factory._socket)

    async def test_exit_none_context(self):
        self.factory._context = None

        await self.factory._exit()

    async def test_close(self):
        self.factory._exit = mock.CoroutineMock()

        await self.factory.close()

        self.factory._exit.assert_called()

    async def test_close_supresses_errors(self):
        self.factory._exit = mock.CoroutineMock(side_effect=AttributeError())

        await self.factory.close()

        self.factory._exit.assert_called()

    async def test_call_socket_creates_socket(self):
        self.factory._enter = mock.CoroutineMock()
        args = [object()]
        kwargs = {"key": "value"}

        await self.factory(*args, **kwargs)

        self.factory._enter.assert_called_with(*args, **kwargs)

    async def test_call_socket_returns_open_socket(self):
        self.factory._enter = mock.CoroutineMock()
        socket = mock.MagicMock()
        socket.closed = False
        self.factory._socket = socket

        result = await self.factory()

        self.assertEqual(result, socket)
        self.factory._enter.assert_not_called()

    async def test_call_socket_creates_new_if_closed(self):
        self.factory._exit = mock.CoroutineMock()
        socket = mock.MagicMock()
        socket.closed = True
        self.factory._socket = socket

        await self.factory()

        self.factory._exit.assert_called()


class TestWebSocketTransport(TestCase):
    def setUp(self):
        self.transport = WebSocketTransport(url="example.com/cometd",
                                            incoming_queue=None,
                                            loop=None)

    def test_connection_type(self):
        self.assertEqual(self.transport.connection_type,
                         ConnectionType.WEBSOCKET)

    async def test_get_socket_for_short_request(self):
        channel = "/test/channel"
        self.assertNotIn(channel, self.transport._long_duration_channels)
        expected_socket = object()
        self.transport._socket_factory_short = mock.CoroutineMock(
            return_value=expected_socket
        )
        headers = object()

        result = await self.transport._get_socket(channel, headers)

        self.assertIs(result, expected_socket)
        self.transport._socket_factory_short.assert_called_with(
            self.transport._url,
            ssl=self.transport.ssl,
            headers=headers
        )

    async def test_get_socket_for_long_duration_request(self):
        channel = MetaChannel.CONNECT
        self.assertIn(channel, self.transport._long_duration_channels)
        expected_socket = object()
        self.transport._socket_factory_long = mock.CoroutineMock(
            return_value=expected_socket
        )
        headers = object()

        result = await self.transport._get_socket(channel, headers)

        self.assertIs(result, expected_socket)
        self.transport._socket_factory_long.assert_called_with(
            self.transport._url,
            ssl=self.transport.ssl,
            headers=headers
        )

    def test_get_lock_for_short_request(self):
        channel = "/test/channel"
        self.assertNotIn(channel, self.transport._long_duration_channels)
        expected_lock = object()
        self.transport._socket_lock_short = expected_lock

        result = self.transport._get_socket_lock(channel)

        self.assertIs(result, expected_lock)

    def test_get_lock_for_long_duration_request(self):
        channel = MetaChannel.CONNECT
        self.assertIn(channel, self.transport._long_duration_channels)
        expected_lock = object()
        self.transport._socket_lock_long = expected_lock

        result = self.transport._get_socket_lock(channel)

        self.assertIs(result, expected_lock)

    async def test_close(self):
        self.transport._socket_factory_short = mock.MagicMock()
        self.transport._socket_factory_short.close = mock.CoroutineMock()
        self.transport._socket_factory_long = mock.MagicMock()
        self.transport._socket_factory_long.close = mock.CoroutineMock()
        self.transport._close_http_session = mock.CoroutineMock()

        await self.transport.close()

        self.transport._socket_factory_short.close.assert_called()
        self.transport._socket_factory_long.close.assert_called()
        self.transport._close_http_session.assert_called()

    async def test_send_socket_payload_short_request(self):
        matching_response = object()
        payload = [object()]
        socket = mock.MagicMock()
        socket.send_json = mock.CoroutineMock()
        response_payload = object()
        response = mock.MagicMock()
        response.json = mock.MagicMock(return_value=response_payload)
        socket.receive = mock.CoroutineMock(return_value=response)
        self.transport._consume_payload = mock.CoroutineMock(
            return_value=matching_response
        )

        await self.transport._send_socket_payload(socket, payload)

        socket.send_json.assert_called_with(payload)
        self.transport._consume_payload.assert_called_with(
            response_payload,
            headers=None,
            find_response_for=payload[0]
        )

    async def test_send_socket_payload_long_duration_request(self):
        matching_response = object()
        payload = [object()]
        socket = mock.MagicMock()
        socket.send_json = mock.CoroutineMock()
        response_payload = object()
        response = mock.MagicMock()
        response.json = mock.MagicMock(return_value=response_payload)
        socket.receive = mock.CoroutineMock(side_effect=[response, response])
        self.transport._consume_payload = mock.CoroutineMock(
            side_effect=[False, matching_response]
        )

        await self.transport._send_socket_payload(socket, payload)

        socket.send_json.assert_called_with(payload)
        self.transport._consume_payload.assert_called_with(
            response_payload,
            headers=None,
            find_response_for=payload[0]
        )

    async def test_send_socket_payload_socket_closed(self):
        matching_response = object()
        payload = [object()]
        socket = mock.MagicMock()
        socket.send_json = mock.CoroutineMock()
        response = mock.MagicMock()
        response.type = WSMsgType.CLOSE
        socket.receive = mock.CoroutineMock(return_value=response)
        self.transport._consume_payload = mock.CoroutineMock(
            return_value=matching_response
        )

        with self.assertRaisesRegex(TransportConnectionClosed,
                                    "Received CLOSE message on the "
                                    "factory."):
            await self.transport._send_socket_payload(socket, payload)

        socket.send_json.assert_called_with(payload)
        self.transport._consume_payload.assert_not_called()

    async def test_send_final_payload(self):
        channel = "channel"
        payload = [dict(channel=channel)]
        lock = mock.MagicMock()
        socket = object()
        response = object()
        self.transport._get_socket = mock.CoroutineMock(return_value=socket)
        self.transport._get_socket_lock = mock.MagicMock(return_value=lock)
        self.transport._send_socket_payload = \
            mock.CoroutineMock(return_value=response)
        headers = object()

        result = await self.transport._send_final_payload(payload,
                                                          headers=headers)

        self.assertEqual(result, response)
        self.transport._get_socket.assert_called_with(channel, headers)
        self.transport._get_socket_lock.assert_called_with(channel)
        lock.__aenter__.assert_called()
        lock.__aexit__.assert_called()
        self.transport._send_socket_payload.assert_called_with(socket,
                                                               payload)

    async def test_send_final_payload_transport_error(self):
        channel = "channel"
        payload = [dict(channel=channel)]
        lock = mock.MagicMock()
        socket = object()
        exception = client_exceptions.ClientError("message")
        self.transport._get_socket = mock.CoroutineMock(return_value=socket)
        self.transport._get_socket_lock = mock.MagicMock(return_value=lock)
        self.transport._send_socket_payload = \
            mock.CoroutineMock(side_effect=exception)
        headers = object()

        with self.assertLogs(WebSocketTransport.__module__,
                             level="DEBUG") as log:
            with self.assertRaisesRegex(TransportError, str(exception)):
                await self.transport._send_final_payload(payload,
                                                         headers=headers)

        log_message = "WARNING:{}:Failed to send payload, {}"\
            .format(WebSocketTransport.__module__, exception)
        self.assertEqual(log.output, [log_message])
        self.transport._get_socket.assert_called_with(channel, headers)
        self.transport._get_socket_lock.assert_called_with(channel)
        lock.__aenter__.assert_called()
        lock.__aexit__.assert_called()
        self.transport._send_socket_payload.assert_called_with(socket,
                                                               payload)

    async def test_send_final_payload_connection_closed_error(self):
        channel = "channel"
        payload = [dict(channel=channel)]
        lock = mock.MagicMock()
        socket = object()
        socket2 = object()
        response = object()
        self.transport._get_socket = mock.CoroutineMock(
            side_effect=[socket, socket2])
        self.transport._get_socket_lock = mock.MagicMock(return_value=lock)
        error = TransportConnectionClosed()
        self.transport._send_socket_payload = \
            mock.CoroutineMock(side_effect=[error, response])
        headers = object()

        result = await self.transport._send_final_payload(payload,
                                                          headers=headers)

        self.assertEqual(result, response)
        self.transport._get_socket.assert_has_calls([
            mock.call(channel, headers), mock.call(channel, headers)])
        self.transport._get_socket_lock.assert_called_with(channel)
        lock.__aenter__.assert_called()
        lock.__aexit__.assert_called()
        self.transport._send_socket_payload.assert_has_calls([
            mock.call(socket, payload), mock.call(socket2, payload)
        ])



"""Client for the CometD Chat Example"""
import asyncio
import argparse

from aioconsole import ainput

from aiocometd import Client, ConnectionType
from aiocometd.exceptions import AiocometdException


async def chat_client(url, nickname, connection_type):
    """Runs the chat client until it's canceled

    :param str url: CometD server URL
    :param str nickname: The user's nickname
    :param aiocometd.ConnectionType connection_type: Connection type
    """
    try:
        room_name = "demo"
        room_channel = "/chat/" + room_name
        members_changed_channel = "/members/" + room_name
        members_channel = "/service/members"

        # start the client with the given connection type
        async with Client(url, connection_type) as client:
            print(f"Connected to '{url}' using connection "
                  f"type '{connection_type.value}'\n")

            # subscribe to the chat room's channel to receive messages
            await client.subscribe(room_channel)

            # subscribe to the members channel to get notifications when the
            # list of the room's members changes
            await client.subscribe(members_changed_channel)

            # publish to the room's channel that the user has joined
            await client.publish(room_channel, {
                "user": nickname,
                "membership": "join",
                "chat": nickname + " has joined"
            })

            # add the user to the room's members
            await client.publish(members_channel, {
                "user": nickname,
                "room": room_channel
            })

            # start the message publisher task
            input_task = asyncio.ensure_future(
                input_publisher(client, nickname, room_channel))

            last_user = None
            try:
                # listen for incoming messages
                async for message in client:
                    # if a chat message is received
                    if message["channel"] == room_channel:
                        data = message["data"]
                        if data["user"] == last_user:
                            user = "..."
                        else:
                            last_user = data["user"]
                            user = data["user"] + ":"
                        # print the incoming message
                        print(f"{user} {data['chat']}")

                    # if the room's members change
                    elif message["channel"] == members_changed_channel:
                        print("MEMBERS:", ", ".join(message["data"]))
                        last_user = None

            finally:
                input_task.cancel()

    except AiocometdException as error:
        print("Encountered an error: " + str(error))
    except asyncio.CancelledError:
        pass
    finally:
        print("\nExiting...")


async def input_publisher(client, nickname, room_channel):
    """Read text from stdin and publish it on the *room_channel*

    :param aiocometd.Client client: A client object
    :param str nickname: The user's nickname
    :param str room_channel: The chat room's channel
    """
    up_one_line = "\033[F"
    clear_line = "\033[K"

    while True:
        try:
            # read from stdin
            message_text = await ainput("")
        except asyncio.CancelledError:
            return

        # clear the last printed line
        print(up_one_line, end="")
        print(clear_line, end="", flush=True)

        # publish the message on the room's channel
        await client.publish(room_channel, {
            "user": nickname,
            "chat": message_text
        })


def get_arguments():
    """Returns the argument's parsed from the command line

    :rtype: dict
    """
    parser = argparse.ArgumentParser(description="CometD chat example client")
    parser.add_argument("url", metavar="server_url", type=str,
                        help="CometD server URL")
    parser.add_argument("nickname", type=str, help="Chat nickname")
    parser.add_argument("-c", "--connection_type", type=ConnectionType,
                        choices=list(ConnectionType),
                        default=ConnectionType.WEBSOCKET.value,
                        help="Connection type")

    return vars(parser.parse_args())


def main():
    """Starts the chat client application"""
    arguments = get_arguments()

    loop = asyncio.get_event_loop()
    chat_task = asyncio.ensure_future(chat_client(**arguments), loop=loop)
    try:
        loop.run_until_complete(chat_task)
    except KeyboardInterrupt:
        chat_task.cancel()
        loop.run_until_complete(chat_task)


if __name__ == "__main__":
    main()



"""Exception types

Exception hierarchy::

    AiocometdException
        ClientError
            ClientInvalidOperation
        TransportError
            TransportInvalidOperation
            TransportTimeoutError
            TransportConnectionClosed
        ServerError
"""
from . import utils


class AiocometdException(Exception):
    """Base exception type.

    All exceptions of the package inherit from this class.
    """


class TransportError(AiocometdException):
    """Error during the transportation of messages"""


class TransportInvalidOperation(TransportError):
    """The requested operation can't be executed on the current state of the
    transport"""


class TransportTimeoutError(TransportError):
    """Transport timeout"""


class TransportConnectionClosed(TransportError):
    """The connection unexpectedly closed"""


class ServerError(AiocometdException):
    """CometD server side error

    If the *response* contains an error field it gets parsed
    according to the \
    `specs <https://docs.cometd.org/current/reference/#_code_error_code>`_

    :param str message: Error description
    :param dict response: Server response message
    """

    @property
    def message(self):
        """Error description"""
        return self.args[0]

    @property
    def response(self):
        """Server response message"""
        return self.args[1]

    @property
    def error(self):
        """Error field in the :obj:`response`"""
        return self.response.get("error")

    @property
    def error_code(self):
        """Error code part of the error code part of the `error\
        <https://docs.cometd.org/current/reference/#_code_error_code>`_, \
        message field"""
        return utils.get_error_code(self.error)

    @property
    def error_message(self):
        """Description part of the `error\
        <https://docs.cometd.org/current/reference/#_code_error_code>`_, \
        message field"""
        return utils.get_error_message(self.error)

    @property
    def error_args(self):
        """Arguments part of the `error\
        <https://docs.cometd.org/current/reference/#_code_error_code>`_, \
        message field"""
        return utils.get_error_args(self.error)


class ClientError(AiocometdException):
    """ComtedD client side error"""


class ClientInvalidOperation(ClientError):
    """The requested operation can't be executed on the current state of the
    client"""



"""Websocket transport class definition"""
import asyncio
import logging
from contextlib import suppress

import aiohttp

from ..constants import ConnectionType, MetaChannel
from .registry import register_transport
from .base import TransportBase
from ..exceptions import TransportError, TransportConnectionClosed


LOGGER = logging.getLogger(__name__)


class WebSocketFactory:  # pylint: disable=too-few-public-methods
    """Helper class to create asynchronous callable objects, that return
    factory objects

    This class allows the usage of factory objects without context blocks
    """
    def __init__(self, session_factory):
        """
        :param asyncio.coroutine session_factory: Coroutine factory function \
        which returns an HTTP session
        """
        self._session_factory = session_factory
        self._context = None
        self._socket = None

    async def close(self):
        """Close the factory"""
        with suppress(Exception):
            await self._exit()

    async def __call__(self, *args, **kwargs):
        """Create a new factory object or returns a previously created one
        if it's not closed

        :param args: positional arguments for the ws_connect function
        :param kwargs: keyword arguments for the ws_connect function
        :return: Websocket object
        :rtype: `aiohttp.ClientWebSocketResponse \
        <https://aiohttp.readthedocs.io/en/stable\
        /client_reference.html#aiohttp.ClientWebSocketResponse>`_
        """
        # if a the factory object already exists and if it's in closed state
        # exit the context manager and clear the references
        if self._socket is not None and self._socket.closed:
            await self._exit()

        # if there is no factory object, then create it and enter the \
        # context manager to initialize it
        if self._socket is None:
            await self._enter(*args, **kwargs)

        return self._socket

    async def _enter(self, *args, **kwargs):
        """Enter factory context

        :param args: positional arguments for the ws_connect function
        :param kwargs: keyword arguments for the ws_connect function
        """
        session = await self._session_factory()
        self._context = session.ws_connect(*args, **kwargs)
        self._socket = await self._context.__aenter__()

    async def _exit(self):
        """Exit factory context"""
        if self._context:
            await self._context.__aexit__(None, None, None)
            self._socket = self._context = None


@register_transport(ConnectionType.WEBSOCKET)
class WebSocketTransport(TransportBase):
    """WebSocket type transport"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        #: channels used during the connect task, requests on these channels
        #: are usually long running
        self._long_duration_channels = (MetaChannel.HANDSHAKE,
                                        MetaChannel.CONNECT)

        #: factory creating sockets for short duration requests
        self._socket_factory_short = WebSocketFactory(self._get_http_session)
        #: factory creating sockets for long duration requests
        self._socket_factory_long = WebSocketFactory(self._get_http_session)

        #: exclusive lock for the objects created by _socket_factory_short
        self._socket_lock_short = asyncio.Lock()
        #: exclusive lock for the objects created by _socket_factory_long
        self._socket_lock_long = asyncio.Lock()

    async def _get_socket(self, channel, headers):
        """Get a websocket object for the given *channel*

        Returns different websocket objects for long running and short duration
        requests, so while a long running request is pending, short duration
        requests can be transmitted.

        :param str channel: CometD channel name
        :param dict headers: Headers to send
        :return: Websocket object
        :rtype: `aiohttp.ClientWebSocketResponse \
        <https://aiohttp.readthedocs.io/en/stable\
        /client_reference.html#aiohttp.ClientWebSocketResponse>`_
        """
        if channel in self._long_duration_channels:
            factory = self._socket_factory_long
        else:
            factory = self._socket_factory_short

        return await factory(self.endpoint, ssl=self.ssl, headers=headers)

    def _get_socket_lock(self, channel):
        """Get an exclusive lock object for the given *channel*

        :param str channel: CometD channel name
        :return: lock object for the *channel*
        :rtype: asyncio.Lock
        """
        if channel in self._long_duration_channels:
            return self._socket_lock_long
        return self._socket_lock_short

    async def _send_final_payload(self, payload, *, headers):
        try:
            # the channel of the first message
            channel = payload[0]["channel"]
            # lock the socket until we're done sending the payload and
            # receiving the response
            lock = self._get_socket_lock(channel)
            async with lock:
                try:
                    # try to send the payload on the socket which might have
                    # been closed since the last time it was used
                    socket = await self._get_socket(channel, headers)
                    return await self._send_socket_payload(socket, payload)
                except TransportConnectionClosed:
                    # if the socket was indeed closed, try to reopen the socket
                    # and send the payload, since the connection could've
                    # normalised since the last network problem
                    socket = await self._get_socket(channel, headers)
                    return await self._send_socket_payload(socket, payload)

        except aiohttp.client_exceptions.ClientError as error:
            LOGGER.warning("Failed to send payload, %s", error)
            raise TransportError(str(error)) from error

    async def _send_socket_payload(self, socket, payload):
        """Send *payload* to the server on the given *socket*

        :param socket: WebSocket object
        :type socket: `aiohttp.ClientWebSocketResponse \
        <https://aiohttp.readthedocs.io/en/stable\
        /client_reference.html#aiohttp.ClientWebSocketResponse>`_
        :param list[dict] payload: A message or a list of messages
        :return: Response payload
        :rtype: list[dict]
        :raises TransportError: When the request fails.
        :raises TransportConnectionClosed: When the *socket* receives a CLOSE \
        message instead of the expected response
        """
        # receive responses from the server and consume them,
        # until we get back the response for the first message in the *payload*
        await socket.send_json(payload)
        while True:
            response = await socket.receive()
            if response.type == aiohttp.WSMsgType.CLOSE:
                raise TransportConnectionClosed("Received CLOSE message on "
                                                "the factory.")
            response_payload = response.json()
            matching_response = await self._consume_payload(
                response_payload,
                headers=None,
                find_response_for=payload[0])
            if matching_response:
                return matching_response

    async def close(self):
        await self._socket_factory_short.close()
        await self._socket_factory_long.close()
        await super().close()



import asyncio
import reprlib
from enum import Enum, unique

from asynctest import TestCase, mock

from aiocometd.client import Client
from aiocometd.exceptions import ServerError, ClientInvalidOperation, \
    TransportError, TransportTimeoutError, ClientError
from aiocometd.constants import DEFAULT_CONNECTION_TYPE, \
    ConnectionType, MetaChannel, SERVICE_CHANNEL_PREFIX, TransportState


@unique
class MockConnectionType(Enum):
    TYPE1 = "type1"
    TYPE2 = "type2"
    TYPE3 = "type3"
    TYPE4 = "type4"


class TestClient(TestCase):
    def setUp(self):
        self.client = Client("")

    async def long_task(self, result, timeout=None):
        if timeout:
            await asyncio.sleep(timeout, loop=self.loop)
        if not isinstance(result, Exception):
            return result
        else:
            raise result

    def test_init_with_loop(self):
        loop = object()

        client = Client(url=None, loop=loop)

        self.assertIs(client._loop, loop)

    @mock.patch("aiocometd.client.asyncio")
    def test_init_without_loop(self, asyncio_mock):
        loop = object()
        asyncio_mock.get_event_loop.return_value = loop

        client = Client(url=None)

        self.assertIs(client._loop, loop)

    def test_init_with_no_connection_types(self):
        client = Client(url=None)

        self.assertEqual(client._connection_types,
                         [ConnectionType.WEBSOCKET,
                          ConnectionType.LONG_POLLING])

    def test_init_with_connection_types_list(self):
        list = [ConnectionType.LONG_POLLING, ConnectionType.WEBSOCKET]

        client = Client(url=None, connection_types=list)

        self.assertEqual(client._connection_types, list)

    def test_init_with_connection_type_value(self):
        type = ConnectionType.LONG_POLLING

        client = Client(url=None, connection_types=type)

        self.assertEqual(client._connection_types, [type])

    def test_subscriptions(self):
        self.client._transport = mock.MagicMock()
        self.client._transport.subscriptions = {"channel1", "channel2"}

        result = self.client.subscriptions

        self.assertEqual(result, self.client._transport.subscriptions)

    def test_subscriptions_emtpy_on_none_transport(self):
        self.client._transport = None

        result = self.client.subscriptions

        self.assertEqual(result, set())

    def test_connection_type(self):
        self.client._transport = mock.MagicMock()
        self.client._transport.connection_type = object()

        result = self.client.connection_type

        self.assertIs(result, self.client._transport.connection_type)

    def test_connection_type_none_on_no_transport(self):
        self.assertIsNone(self.client.connection_type)

    def test_closed(self):
        self.assertIs(self.client.closed, self.client._closed)

    def test_closed_read_only(self):
        with self.assertRaises(AttributeError):
            self.client.closed = False

    @mock.patch("aiocometd.client.ConnectionType", new=MockConnectionType)
    def test_pick_connection_type(self):
        self.client._connection_types = [
            MockConnectionType.TYPE1,
            MockConnectionType.TYPE2,
            MockConnectionType.TYPE3
        ]
        supported_types = [
            MockConnectionType.TYPE2.value,
            MockConnectionType.TYPE3.value,
            MockConnectionType.TYPE4.value
        ]

        result = self.client._pick_connection_type(supported_types)

        self.assertEqual(result, MockConnectionType.TYPE2)

    @mock.patch("aiocometd.client.ConnectionType", new=MockConnectionType)
    def test_pick_connection_type_without_overlap(self):
        self.client._connection_types = [
            MockConnectionType.TYPE1,
            MockConnectionType.TYPE2
        ]
        supported_types = [
            MockConnectionType.TYPE3.value,
            MockConnectionType.TYPE4.value
        ]

        result = self.client._pick_connection_type(supported_types)

        self.assertIsNone(result)

    @mock.patch("aiocometd.client.create_transport")
    async def test_negotiate_transport_default(self, create_transport):
        response = {
            "supportedConnectionTypes": [DEFAULT_CONNECTION_TYPE.value],
            "successful": True
        }
        transport = mock.MagicMock()
        transport.connection_type = DEFAULT_CONNECTION_TYPE
        transport.handshake = mock.CoroutineMock(return_value=response)
        create_transport.return_value = transport
        self.client._pick_connection_type = \
            mock.MagicMock(return_value=DEFAULT_CONNECTION_TYPE)
        self.client._verify_response = mock.MagicMock()
        self.client.extensions = object()
        self.client.auth = object()

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            result = await self.client._negotiate_transport()

        self.assertEqual(result, transport)
        create_transport.assert_called_with(
            DEFAULT_CONNECTION_TYPE,
            url=self.client.url,
            incoming_queue=self.client._incoming_queue,
            ssl=self.client.ssl,
            extensions=self.client.extensions,
            auth=self.client.auth,
            loop=self.client._loop)
        transport.handshake.assert_called_with(self.client._connection_types)
        self.client._verify_response.assert_called_with(response)
        self.client._pick_connection_type.assert_called_with(
            response["supportedConnectionTypes"])
        log_message = ("INFO:aiocometd.client:"
                       "Connection types supported by the server: {!r}"
                       .format(response["supportedConnectionTypes"]))
        self.assertEqual(log.output, [log_message])

    @mock.patch("aiocometd.client.create_transport")
    async def test_negotiate_transport_error(self, create_transport):
        response = {
            "supportedConnectionTypes": [DEFAULT_CONNECTION_TYPE.value],
            "successful": True
        }
        transport = mock.MagicMock()
        transport.connection_type = DEFAULT_CONNECTION_TYPE
        transport.handshake = mock.CoroutineMock(return_value=response)
        transport.close = mock.CoroutineMock()
        create_transport.return_value = transport
        self.client._pick_connection_type = \
            mock.MagicMock(return_value=None)
        self.client.extensions = object()
        self.client.auth = object()

        with self.assertRaisesRegex(ClientError,
                                    "None of the connection types offered "
                                    "by the server are supported."):
            with self.assertLogs("aiocometd.client", "DEBUG") as log:
                await self.client._negotiate_transport()

        create_transport.assert_called_with(
            DEFAULT_CONNECTION_TYPE,
            url=self.client.url,
            incoming_queue=self.client._incoming_queue,
            ssl=self.client.ssl,
            extensions=self.client.extensions,
            auth=self.client.auth,
            loop=self.client._loop)
        transport.handshake.assert_called_with(self.client._connection_types)
        self.client._pick_connection_type.assert_called_with(
            response["supportedConnectionTypes"])
        transport.close.assert_called()
        log_message = ("INFO:aiocometd.client:"
                       "Connection types supported by the server: {!r}"
                       .format(response["supportedConnectionTypes"]))
        self.assertEqual(log.output, [log_message])

    @mock.patch("aiocometd.client.create_transport")
    async def test_negotiate_transport_non_default(self, create_transport):
        non_default_type = ConnectionType.WEBSOCKET
        self.client._connection_types = [non_default_type]
        response = {
            "supportedConnectionTypes": [DEFAULT_CONNECTION_TYPE.value,
                                         non_default_type.value],
            "successful": True
        }
        transport1 = mock.MagicMock()
        transport1.connection_type = DEFAULT_CONNECTION_TYPE
        transport1.client_id = "client_id"
        transport1.handshake = mock.CoroutineMock(return_value=response)
        transport1.close = mock.CoroutineMock()
        transport2 = mock.MagicMock()
        transport2.connection_type = non_default_type
        transport2.client_id = None
        create_transport.side_effect = [transport1, transport2]
        self.client._pick_connection_type = \
            mock.MagicMock(return_value=non_default_type)
        self.client._verify_response = mock.MagicMock()
        self.client.extensions = object()
        self.client.auth = object()

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            result = await self.client._negotiate_transport()

        self.assertEqual(result, transport2)
        create_transport.assert_has_calls(
            [
                mock.call(
                    DEFAULT_CONNECTION_TYPE,
                    url=self.client.url,
                    incoming_queue=self.client._incoming_queue,
                    ssl=self.client.ssl,
                    extensions=self.client.extensions,
                    auth=self.client.auth,
                    loop=self.client._loop),
                mock.call(
                    non_default_type,
                    url=self.client.url,
                    incoming_queue=self.client._incoming_queue,
                    client_id=transport1.client_id,
                    ssl=self.client.ssl,
                    extensions=self.client.extensions,
                    auth=self.client.auth,
                    loop=self.client._loop)
            ]
        )
        transport1.handshake.assert_called_with(self.client._connection_types)
        self.client._verify_response.assert_called_with(response)
        self.client._pick_connection_type.assert_called_with(
            response["supportedConnectionTypes"])
        transport1.close.assert_called()
        log_message = ("INFO:aiocometd.client:"
                       "Connection types supported by the server: {!r}"
                       .format(response["supportedConnectionTypes"]))
        self.assertEqual(log.output, [log_message])

    async def test_open(self):
        transport = mock.MagicMock()
        transport.connection_type = ConnectionType.LONG_POLLING
        connect_result = object()
        transport.connect = mock.CoroutineMock(return_value=connect_result)
        self.client._negotiate_transport = mock.CoroutineMock(
            return_value=transport
        )
        self.client._verify_response = mock.MagicMock()
        self.client._closed = True

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            await self.client.open()

        self.client._negotiate_transport.assert_called()
        transport.connect.assert_called()
        self.client._verify_response.assert_called_with(connect_result)
        self.assertEqual(
            log.output,
            ["INFO:aiocometd.client:Opening client with connection "
             "types {!r} ..."
             .format([t.value for t in self.client._connection_types]),
             "INFO:aiocometd.client:Client opened with connection_type {!r}"
             .format(self.client.connection_type.value)])

    async def test_open_if_already_open(self):
        transport = mock.MagicMock()
        connect_result = object()
        transport.connect = mock.CoroutineMock(return_value=connect_result)
        self.client._negotiate_transport = mock.CoroutineMock(
            return_value=transport
        )
        self.client._verify_response = mock.MagicMock()
        self.client._closed = False

        with self.assertRaisesRegex(ClientInvalidOperation,
                                    "Client is already open."):
            await self.client.open()

        self.client._negotiate_transport.assert_not_called()
        transport.connect.assert_not_called()
        self.client._verify_response.assert_not_called()

    async def test_close(self):
        self.client._closed = False
        self.client._transport = mock.MagicMock()
        self.client._transport.client_id = "client_id"
        self.client._transport.disconnect = mock.CoroutineMock()
        self.client._transport.close = mock.CoroutineMock()
        expected_log = [
            "INFO:aiocometd.client:Closing client...",
            "INFO:aiocometd.client:Client closed."
        ]

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            await self.client.close()

        self.client._transport.disconnect.assert_called()
        self.client._transport.close.assert_called()
        self.assertTrue(self.client.closed)
        self.assertEqual(log.output, expected_log)

    async def test_close_with_pending_messages(self):
        self.client._closed = False
        self.client._transport = mock.MagicMock()
        self.client._transport.client_id = "client_id"
        self.client._transport.disconnect = mock.CoroutineMock()
        self.client._transport.close = mock.CoroutineMock()
        self.client._incoming_queue = asyncio.Queue()
        self.client._incoming_queue.put_nowait(object())
        expected_log = [
            "WARNING:aiocometd.client:Closing client while {} messages are "
            "still pending...".format(self.client.pending_count),
            "INFO:aiocometd.client:Client closed."
        ]

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            await self.client.close()

        self.client._transport.disconnect.assert_called()
        self.client._transport.close.assert_called()
        self.assertTrue(self.client.closed)
        self.assertEqual(log.output, expected_log)

    async def test_close_if_already_closed(self):
        self.client._closed = True
        self.client._transport = mock.MagicMock()
        self.client._transport.client_id = "client_id"
        self.client._transport.disconnect = mock.CoroutineMock()
        self.client._transport.close = mock.CoroutineMock()

        await self.client.close()

        self.client._transport.disconnect.assert_not_called()
        self.client._transport.close.assert_not_called()
        self.assertTrue(self.client.closed)

    async def test_close_on_transport_error(self):
        self.client._closed = False
        self.client._transport = mock.MagicMock()
        self.client._transport.client_id = "client_id"
        error = TransportError("description")
        self.client._transport.disconnect = mock.CoroutineMock(
            side_effect=error
        )
        self.client._transport.close = mock.CoroutineMock()
        expected_log = ["INFO:aiocometd.client:Closing client...",
                        "INFO:aiocometd.client:Client closed."]

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            with self.assertRaisesRegex(TransportError, str(error)):
                await self.client.close()

        self.assertEqual(log.output, expected_log)
        self.client._transport.disconnect.assert_called()
        self.client._transport.close.assert_not_called()
        self.assertTrue(self.client.closed)

    async def test_close_no_transport(self):
        self.client._closed = False
        self.client._transport = None
        expected_log = [
            "INFO:aiocometd.client:Closing client...",
            "INFO:aiocometd.client:Client closed."
        ]

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            await self.client.close()

        self.assertTrue(self.client.closed)
        self.assertEqual(log.output, expected_log)

    async def test_subscribe(self):
        response = {
            "channel": MetaChannel.SUBSCRIBE,
            "successful": True,
            "subscription": "channel1",
            "id": "1"
        }
        self.client._transport = mock.MagicMock()
        self.client._transport.subscribe = mock.CoroutineMock(
            return_value=response
        )
        self.client._check_server_disconnected = mock.CoroutineMock()
        self.client._closed = False

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            await self.client.subscribe("channel1")

        self.client._transport.subscribe.assert_called_with("channel1")
        self.assertEqual(log.output,
                         ["INFO:aiocometd.client:Subscribed to channel {}"
                          .format("channel1")])
        self.client._check_server_disconnected.assert_called()

    async def test_subscribe_on_closed(self):
        self.client._closed = True
        self.client._check_server_disconnected = mock.CoroutineMock()

        with self.assertRaisesRegex(ClientInvalidOperation,
                                    "Can't send subscribe request while, "
                                    "the client is closed."):
            await self.client.subscribe("channel1")

        self.client._check_server_disconnected.assert_not_called()

    async def test_subscribe_error(self):
        response = {
            "channel": MetaChannel.SUBSCRIBE,
            "successful": False,
            "subscription": "channel1",
            "id": "1"
        }
        self.client._transport = mock.MagicMock()
        self.client._transport.subscribe = mock.CoroutineMock(
            return_value=response
        )
        self.client._check_server_disconnected = mock.CoroutineMock()
        self.client._closed = False
        error = ServerError("Subscribe request failed.", response)

        with self.assertRaisesRegex(ServerError, str(error)):
            await self.client.subscribe("channel1")

        self.client._transport.subscribe.assert_called_with("channel1")
        self.client._check_server_disconnected.assert_called()

    async def test_unsubscribe(self):
        response = {
            "channel": MetaChannel.UNSUBSCRIBE,
            "successful": True,
            "subscription": "channel1",
            "id": "1"
        }
        self.client._transport = mock.MagicMock()
        self.client._transport.unsubscribe = mock.CoroutineMock(
            return_value=response
        )
        self.client._check_server_disconnected = mock.CoroutineMock()
        self.client._closed = False

        with self.assertLogs("aiocometd.client", "DEBUG") as log:
            await self.client.unsubscribe("channel1")

        self.client._transport.unsubscribe.assert_called_with("channel1")
        self.assertEqual(log.output,
                         ["INFO:aiocometd.client:Unsubscribed from channel {}"
                            .format("channel1")])
        self.client._check_server_disconnected.assert_called()

    async def test_unsubscribe_on_closed(self):
        self.client._closed = True
        self.client._check_server_disconnected = mock.CoroutineMock()

        with self.assertRaisesRegex(ClientInvalidOperation,
                                    "Can't send unsubscribe request while, "
                                    "the client is closed."):
            await self.client.unsubscribe("channel1")

        self.client._check_server_disconnected.assert_not_called()

    async def test_unsubscribe_error(self):
        response = {
            "channel": MetaChannel.UNSUBSCRIBE,
            "successful": False,
            "subscription": "channel1",
            "id": "1"
        }
        self.client._transport = mock.MagicMock()
        self.client._transport.unsubscribe = mock.CoroutineMock(
            return_value=response
        )
        self.client._check_server_disconnected = mock.CoroutineMock()
        self.client._closed = False
        error = ServerError("Unsubscribe request failed.", response)

        with self.assertRaisesRegex(ServerError, str(error)):
            await self.client.unsubscribe("channel1")

        self.client._transport.unsubscribe.assert_called_with("channel1")
        self.client._check_server_disconnected.assert_called()

    async def test_publish(self):
        response = {
            "channel": "/channel1",
            "successful": True,
            "id": "1"
        }
        data = {}
        self.client._transport = mock.MagicMock()
        self.client._transport.publish = mock.CoroutineMock(
            return_value=response
        )
        self.client._check_server_disconnected = mock.CoroutineMock()
        self.client._closed = False

        result = await self.client.publish("channel1", data)

        self.assertEqual(result, response)
        self.client._transport.publish.assert_called_with("channel1", data)
        self.client._check_server_disconnected.assert_called()

    async def test_publish_on_closed(self):
        self.client._closed = True
        self.client._check_server_disconnected = mock.CoroutineMock()

        with self.assertRaisesRegex(ClientInvalidOperation,
                                    "Can't publish data while, "
                                    "the client is closed."):
            await self.client.publish("channel1", {})

        self.client._check_server_disconnected.assert_not_called()

    async def test_publish_error(self):
        response = {
            "channel": "/channel1",
            "successful": False,
            "id": "1"
        }
        data = {}
        self.client._transport = mock.MagicMock()
        self.client._transport.publish = mock.CoroutineMock(
            return_value=response
        )
        self.client._check_server_disconnected = mock.CoroutineMock()
        self.client._closed = False
        error = ServerError("Publish request failed.", response)

        with self.assertRaisesRegex(ServerError, str(error)):
            await self.client.publish("channel1", data)

        self.client._transport.publish.assert_called_with("channel1", data)
        self.client._check_server_disconnected.assert_called()

    def test_repr(self):
        self.client.url = "http://example.com"
        expected = "Client({}, {}, connection_timeout={}, ssl={}, " \
                   "max_pending_count={}, extensions={}, auth={}, " \
                   "loop={})".format(
                        reprlib.repr(self.client.url),
                        reprlib.repr(self.client._connection_types),
                        reprlib.repr(self.client.connection_timeout),
                        reprlib.repr(self.client.ssl),
                        reprlib.repr(self.client._max_pending_count),
                        reprlib.repr(self.client.extensions),
                        reprlib.repr(self.client.auth),
                        reprlib.repr(self.client._loop))

        result = repr(self.client)

        self.assertEqual(result, expected)

    def test_verify_response_on_success(self):
        self.client._raise_server_error = mock.MagicMock()
        response = {
            "channel": "/channel1",
            "successful": True,
            "id": "1"
        }

        self.client._verify_response(response)

        self.client._raise_server_error.assert_not_called()

    def test_verify_response_on_error(self):
        self.client._raise_server_error = mock.MagicMock()
        response = {
            "channel": "/channel1",
            "successful": False,
            "id": "1"
        }

        self.client._verify_response(response)

        self.client._raise_server_error.assert_called_with(response)

    def test_verify_response_no_successful_status(self):
        self.client._raise_server_error = mock.MagicMock()
        response = {
            "channel": "/channel1",
            "id": "1"
        }

        self.client._verify_response(response)

        self.client._raise_server_error.assert_not_called()

    def test_raise_server_error_meta(self):
        response = {
            "channel": MetaChannel.SUBSCRIBE,
            "successful": False,
            "id": "1"
        }
        error_message = \
            type(self.client)._SERVER_ERROR_MESSAGES[response["channel"]]

        with self.assertRaisesRegex(ServerError, error_message):
            self.client._raise_server_error(response)

    def test_raise_server_error_service(self):
        response = {
            "channel": SERVICE_CHANNEL_PREFIX + "test",
            "successful": False,
            "id": "1"
        }

        with self.assertRaisesRegex(ServerError, "Service request failed."):
            self.client._raise_server_error(response)

    def test_raise_server_error_publish(self):
        response = {
            "channel": "/some/channel",
            "successful": False,
            "id": "1"
        }

        with self.assertRaisesRegex(ServerError, "Publish request failed."):
            self.client._raise_server_error(response)

    async def test_pending_count(self):
        self.client._incoming_queue = asyncio.Queue()
        await self.client._incoming_queue.put(1)
        await self.client._incoming_queue.put(2)

        self.assertEqual(self.client.pending_count, 2)

    async def test_pending_count_if_none_queue(self):
        self.client._incoming_queue = None

        self.assertEqual(self.client.pending_count, 0)

    async def test_has_pending_messages(self):
        self.client._incoming_queue = asyncio.Queue()
        await self.client._incoming_queue.put(1)

        self.assertTrue(self.client.has_pending_messages)

    async def test_has_pending_messages_false(self):
        self.client._incoming_queue = None

        self.assertFalse(self.client.has_pending_messages)

    async def test_receive_on_closed(self):
        self.client._closed = True
        self.client._incoming_queue = None

        with self.assertRaisesRegex(ClientInvalidOperation,
                                    "The client is closed and there are "
                                    "no pending messages."):
            await self.client.receive()

    async def test_receive_on_closed_and_pending_messages(self):
        self.client._closed = True
        response = {
            "channel": "/channel1",
            "data": {},
            "id": "1"
        }
        self.client._incoming_queue = mock.MagicMock()
        self.client._incoming_queue.qsize.return_value = 1
        self.client._get_message = mock.CoroutineMock(return_value=response)
        self.client._verify_response = mock.MagicMock()

        result = await self.client.receive()

        self.assertEqual(result, response)
        self.client._get_message.assert_called_with(
            self.client.connection_timeout)
        self.client._verify_response.assert_called_with(response)

    async def test_receive_on_open(self):
        self.client._closed = False
        response = {
            "channel": "/channel1",
            "data": {},
            "id": "1"
        }
        self.client._incoming_queue = mock.MagicMock()
        self.client._incoming_queue.qsize.return_value = 1
        self.client._get_message = mock.CoroutineMock(return_value=response)
        self.client._verify_response = mock.MagicMock()

        result = await self.client.receive()

        self.assertEqual(result, response)
        self.client._get_message.assert_called_with(
            self.client.connection_timeout)
        self.client._verify_response.assert_called_with(response)

    async def test_receive_on_connection_timeout(self):
        self.client._closed = True
        self.client._incoming_queue = mock.MagicMock()
        self.client._incoming_queue.qsize.return_value = 1
        self.client._get_message = mock.CoroutineMock(
            side_effect=TransportTimeoutError())
        self.client._verify_response = mock.MagicMock()

        with self.assertRaises(TransportTimeoutError):
            await self.client.receive()

        self.client._get_message.assert_called_with(
            self.client.connection_timeout)
        self.client._verify_response.assert_not_called()

    async def test_aiter(self):
        responses = [
            {
                "channel": "/channel1",
                "data": {},
                "id": "1"
            },
            {
                "channel": "/channel2",
                "data": {},
                "id": "2"
            }
        ]
        self.client.receive = mock.CoroutineMock(
            side_effect=responses + [ClientInvalidOperation()])

        result = []
        async for message in self.client:
            result.append(message)

        self.assertEqual(result, responses)

    async def test_context_manager(self):
        self.client.open = mock.CoroutineMock()
        self.client.close = mock.CoroutineMock()

        async with self.client as client:
            pass

        self.assertIs(client, self.client)
        self.client.open.assert_called()
        self.client.close.assert_called()

    async def test_context_manager_on_enter_error(self):
        self.client.open = mock.CoroutineMock(
            side_effect=TransportError()
        )
        self.client.close = mock.CoroutineMock()

        with self.assertRaises(TransportError):
            async with self.client:
                pass

        self.client.open.assert_called()
        self.client.close.assert_called()

    @mock.patch("aiocometd.client.asyncio")
    async def test_wait_connection_timeout_on_timeout(self, asyncio_mock):
        self.client._transport = mock.MagicMock()
        self.client._transport.wait_for_state = mock.CoroutineMock()
        timeout = 2
        asyncio_mock.wait_for = mock.CoroutineMock(
            side_effect=asyncio.TimeoutError()
        )
        asyncio_mock.TimeoutError = asyncio.TimeoutError

        await self.client._wait_connection_timeout(timeout)

        self.client._transport.wait_for_state.assert_has_calls([
            mock.call(TransportState.CONNECTING),
            mock.call(TransportState.CONNECTED)
        ])
        asyncio_mock.wait_for.assert_called()

    @mock.patch("aiocometd.client.asyncio")
    async def test_wait_connection_timeout_iterations(self, asyncio_mock):
        self.client._transport = mock.MagicMock()
        self.client._transport.wait_for_state = mock.CoroutineMock()
        timeout = 2
        asyncio_mock.wait_for = mock.CoroutineMock(
            side_effect=[None, asyncio.TimeoutError()]
        )
        asyncio_mock.TimeoutError = asyncio.TimeoutError

        await self.client._wait_connection_timeout(timeout)

        self.client._transport.wait_for_state.assert_has_calls([
            mock.call(TransportState.CONNECTING),
            mock.call(TransportState.CONNECTED),
            mock.call(TransportState.CONNECTING),
            mock.call(TransportState.CONNECTED)
        ])
        asyncio_mock.wait_for.assert_called()

    async def test_get_message_no_timeout(self):
        self.client._incoming_queue = mock.MagicMock()
        self.client._incoming_queue.get = mock.CoroutineMock(
            return_value=object()
        )
        self.client._wait_connection_timeout = mock.MagicMock()
        self.client._transport = mock.MagicMock()
        self.client._transport.wait_for_state = mock.MagicMock(
            return_value=self.long_task(None, timeout=1)
        )

        result = await self.client._get_message(None)

        self.assertIs(result, self.client._incoming_queue.get.return_value)
        self.client._wait_connection_timeout.assert_not_called()
        self.client._transport.wait_for_state.assert_called_with(
            TransportState.SERVER_DISCONNECTED
        )

    async def test_get_message_with_timeout_not_triggered(self):
        self.client._incoming_queue = mock.MagicMock()
        get_result = object()
        self.client._incoming_queue.get = mock.MagicMock(
            return_value=self.long_task(get_result, timeout=None)
        )
        self.client._wait_connection_timeout = mock.MagicMock(
            return_value=self.long_task(None, timeout=1)
        )
        self.client._transport = mock.MagicMock()
        self.client._transport.wait_for_state = mock.MagicMock(
            return_value=self.long_task(None, timeout=1)
        )
        timeout = 2

        result = await self.client._get_message(timeout)

        self.assertIs(result, get_result)
        self.client._wait_connection_timeout.assert_called_with(timeout)
        self.client._transport.wait_for_state.assert_called_with(
            TransportState.SERVER_DISCONNECTED
        )

    async def test_get_message_with_timeout_triggered(self):
        self.client._incoming_queue = mock.MagicMock()
        get_result = object()
        self.client._incoming_queue.get = mock.MagicMock(
            return_value=self.long_task(get_result, timeout=1)
        )
        self.client._wait_connection_timeout = mock.MagicMock(
            return_value=self.long_task(None, timeout=None)
        )
        self.client._transport = mock.MagicMock()
        self.client._transport.wait_for_state = mock.MagicMock(
            return_value=self.long_task(None, timeout=1)
        )
        timeout = 2

        with self.assertRaisesRegex(TransportTimeoutError,
                                    "Lost connection with the server"):
            await self.client._get_message(timeout)

        self.client._incoming_queue.get.assert_called()
        self.client._wait_connection_timeout.assert_called_with(timeout)
        self.client._transport.wait_for_state.assert_called_with(
            TransportState.SERVER_DISCONNECTED
        )

    async def test_get_message_with_server_disconnected(self):
        self.client._incoming_queue = mock.MagicMock()
        get_result = object()
        self.client._incoming_queue.get = mock.MagicMock(
            return_value=self.long_task(get_result, timeout=1)
        )
        self.client._wait_connection_timeout = mock.MagicMock(
            return_value=self.long_task(None, timeout=1)
        )
        self.client._transport = mock.MagicMock()
        self.client.close = mock.CoroutineMock()
        self.client._transport.wait_for_state = mock.MagicMock(
            return_value=self.long_task(None, timeout=None)
        )
        timeout = 2
        self.client._transport.state = TransportState.SERVER_DISCONNECTED

        with self.assertRaisesRegex(ServerError,
                                    "Connection closed by the server"):
            await self.client._get_message(timeout)

        self.client._wait_connection_timeout.assert_called_with(timeout)
        self.client._transport.wait_for_state.assert_called_with(
            TransportState.SERVER_DISCONNECTED
        )
        self.client.close.assert_called()

    @mock.patch("aiocometd.client.asyncio")
    async def test_get_message_cancelled(self, asyncio_mock):
        self.client._incoming_queue = mock.MagicMock()
        self.client._wait_connection_timeout = mock.MagicMock()
        self.client._transport = mock.MagicMock()
        self.client._transport.wait_for_state = mock.MagicMock()
        asyncio_mock.ensure_future = mock.MagicMock(
            side_effect=[mock.MagicMock(), mock.MagicMock(), mock.MagicMock()]
        )
        asyncio_mock.wait = mock.CoroutineMock(
            side_effect=asyncio.CancelledError()
        )
        asyncio_mock.CancelledError = asyncio.CancelledError

        with self.assertRaises(asyncio.CancelledError):
            await self.client._get_message(1)

        for task in asyncio_mock.ensure_future.side_effect:
            task.cancel.assert_called()

    async def test_check_server_disconnected_on_disconnected(self):
        self.client._transport = mock.MagicMock()
        self.client._transport.state = TransportState.SERVER_DISCONNECTED
        self.client._transport.last_connect_result = object()

        with self.assertRaisesRegex(ServerError,
                                    "Connection closed by the server") as cm:
            await self.client._check_server_disconnected()

        self.assertEqual(cm.exception.response,
                         self.client._transport.last_connect_result)

    async def test_check_server_disconnected_on_not_disconnected(self):
        self.client._transport = mock.MagicMock()
        self.client._transport.state = TransportState.CONNECTED

        await self.client._check_server_disconnected()

    async def test_check_server_disconnected_on_none_transport(self):
        self.client._transport = None

        await self.client._check_server_disconnected()



"""Base transtport abstract class definition"""
import asyncio
import logging
from abc import abstractmethod
from contextlib import suppress

import aiohttp

from .abc import Transport
from ..constants import MetaChannel, TransportState, HANDSHAKE_MESSAGE, \
    CONNECT_MESSAGE, DISCONNECT_MESSAGE, SUBSCRIBE_MESSAGE, \
    UNSUBSCRIBE_MESSAGE, PUBLISH_MESSAGE
from ..utils import defer, is_matching_response, is_auth_error_message, \
    is_server_error_message, is_event_message
from ..exceptions import TransportInvalidOperation, TransportError


LOGGER = logging.getLogger(__name__)


class TransportBase(Transport):  # pylint: disable=too-many-instance-attributes
    """Base transport implementation

    This class contains most of the transport operations implemented, it can
    be used as a base class for various concrete transport implementations.
    When subclassing, at a minimum the :meth:`_send_final_payload` and
    :obj:`~Transport.connection_type` methods should be reimplemented.
    """
    #: Timeout to give to HTTP session to close itself
    _HTTP_SESSION_CLOSE_TIMEOUT = 0.250

    def __init__(self, *, url, incoming_queue, client_id=None,
                 reconnection_timeout=1, ssl=None, extensions=None, auth=None,
                 loop=None):
        """
        :param str url: CometD service url
        :param asyncio.Queue incoming_queue: Queue for consuming incoming event
                                             messages
        :param str client_id: Clinet id value assigned by the server
        :param reconnection_timeout: The time to wait before trying to \
        reconnect to the server after a network failure
        :type reconnection_timeout: None or int or float
        :param ssl: SSL validation mode. None for default SSL check \
        (:func:`ssl.create_default_context` is used), False for skip SSL \
        certificate validation, \
        `aiohttp.Fingerprint <https://aiohttp.readthedocs.io/en/stable/\
        client_reference.html#aiohttp.Fingerprint>`_ for fingerprint \
        validation, :obj:`ssl.SSLContext` for custom SSL certificate \
        validation.
        :param extensions: List of protocol extension objects
        :type extensions: list[Extension] or None
        :param AuthExtension auth: An auth extension
        :param loop: Event :obj:`loop <asyncio.BaseEventLoop>` used to
                     schedule tasks. If *loop* is ``None`` then
                     :func:`asyncio.get_event_loop` is used to get the default
                     event loop.
        """
        #: queue for consuming incoming event messages
        self.incoming_queue = incoming_queue
        #: CometD service url
        self._url = url
        #: event loop used to schedule tasks
        self._loop = loop or asyncio.get_event_loop()
        #: clinet id value assigned by the server
        self._client_id = client_id
        #: message id which should be unique for every message during a client
        #: session
        self._message_id = 0
        #: reconnection advice parameters returned by the server
        self._reconnect_advice = {}
        #: set of subscribed channels
        self._subscriptions = set()
        #: boolean to mark whether to resubscribe on connect
        self._subscribe_on_connect = False
        #: dictionary of TransportState and asyncio.Event pairs
        self._state_events = {_: asyncio.Event() for _ in TransportState}
        #: current state of the transport
        self._state = TransportState.DISCONNECTED
        #: asyncio connection task
        self._connect_task = None
        #: time to wait before reconnecting after a network failure
        self._reconnect_timeout = reconnection_timeout
        #: SSL validation mode
        self.ssl = ssl
        #: http session
        self._http_session = None
        #: List of protocol extension objects
        self._extensions = extensions or []
        #: An auth extension
        self._auth = auth

    async def _get_http_session(self):
        """Factory method for getting the current HTTP session

        :return: The current session if it's not None, otherwise it creates a
                 new session.
        """
        # it would be nicer to create the session when the class gets
        # initialized, but this seems to be the right way to do it since
        # aiohttp produces log messages with warnings that a session should be
        # created in a coroutine
        if self._http_session is None:
            self._http_session = aiohttp.ClientSession()
        return self._http_session

    async def _close_http_session(self):
        """Close the http session if it's not already closed"""
        # graceful shutdown recommended by the documentation
        # https://aiohttp.readthedocs.io/en/stable/client_advanced.html\
        # #graceful-shutdown
        if self._http_session is not None and not self._http_session.closed:
            await self._http_session.close()
            await asyncio.sleep(self._HTTP_SESSION_CLOSE_TIMEOUT)

    @property
    def connection_type(self):
        """The transport's connection type"""
        return None  # pragma: no cover

    @property
    def endpoint(self):
        """CometD service url"""
        return self._url

    @property
    def client_id(self):
        """clinet id value assigned by the server"""
        return self._client_id

    @property
    def subscriptions(self):
        """Set of subscribed channels"""
        return self._subscriptions

    @property
    def last_connect_result(self):
        """Result of the last connect request"""
        if self._connect_task and self._connect_task.done():
            return self._connect_task.result()
        return None

    @property
    def state(self):
        """Current state of the transport"""
        return self._state

    @property
    def _state(self):
        """Current state of the transport"""
        return self.__dict__.setdefault("_state")

    @_state.setter
    def _state(self, value):
        self._set_state_event(self._state, value)
        self.__dict__["_state"] = value

    def _set_state_event(self, old_state, new_state):
        """Set event associated with the *new_state* and clear the event for
        the *old_state*

        :param old_state: Old state value
        :type old_state: TransportState or None
        :param new_state: New state value
        :type new_state: TransportState
        """
        if new_state != old_state:
            if old_state is not None:
                self._state_events[old_state].clear()
            self._state_events[new_state].set()

    async def wait_for_state(self, state):
        """Waits for and returns when the transport enters the given *state*

        :param TransportState state: A state value
        """
        await self._state_events[state].wait()

    async def handshake(self, connection_types):
        """Executes the handshake operation

        :param list[ConnectionType] connection_types: list of connection types
        :return: Handshake response
        :rtype: dict
        :raises TransportError: When the network request fails.
        """
        # reset message id for a new client session
        self._message_id = 0

        connection_types = list(connection_types)
        # make sure that the supported connection types list contains this
        # transport
        if self.connection_type not in connection_types:
            connection_types.append(self.connection_type)
        connection_type_values = [ct.value for ct in connection_types]

        # send message and await its response
        response_message = await self._send_message(
            HANDSHAKE_MESSAGE.copy(),
            supportedConnectionTypes=connection_type_values
        )
        # store the returned client id or set it to None if it's not in the
        # response
        if response_message["successful"]:
            self._client_id = response_message.get("clientId")
            self._subscribe_on_connect = True
        return response_message

    def _finalize_message(self, message):
        """Update the ``id``, ``clientId`` and ``connectionType`` message
        fields as a side effect if they're are present in the *message*.

        :param dict message: Outgoing message
        """
        if "id" in message:
            message["id"] = str(self._message_id)
            self._message_id += 1

        if "clientId" in message:
            message["clientId"] = self.client_id

        if "connectionType" in message:
            message["connectionType"] = self.connection_type.value

    def _finalize_payload(self, payload):
        """Update the ``id``, ``clientId`` and ``connectionType`` message
        fields in the *payload*, as a side effect if they're are present in
        the *message*. The *payload* can be either a single message or a list
        of messages.

        :param payload: A message or a list of messages
        :type payload: dict or list[dict]
        """
        if isinstance(payload, list):
            for item in payload:
                self._finalize_message(item)
        else:
            self._finalize_message(payload)

    async def _send_message(self, message, **kwargs):
        """Send message to server

        :param dict message: A message
        :param kwargs: Optional key-value pairs that'll be used to update the \
        the values in the *message*
        :return: Response message
        :rtype: dict
        :raises TransportError: When the network request fails.
        """
        message.update(kwargs)
        return await self._send_payload_with_auth([message])

    async def _send_payload_with_auth(self, payload):
        """Finalize and send *payload* to server and retry on authentication
        failure

        Finalize and send the *payload* to the server and return once a
        response message can be provided for the first message in the
        *payload*.

        :param list[dict] payload: A list of messages
        :return: The response message for the first message in the *payload*
        :rtype: dict
        :raises TransportError: When the network request fails.
        """
        response = await self._send_payload(payload)

        # if there is an auth extension and the response is an auth error
        if self._auth and is_auth_error_message(response):
            # then try to authenticate and resend the payload
            await self._auth.authenticate()
            return await self._send_payload(payload)

        # otherwise return the response
        return response

    async def _send_payload(self, payload):
        """Finalize and send *payload* to server

        Finalize and send the *payload* to the server and return once a
        response message can be provided for the first message in the
        *payload*.

        :param list[dict] payload: A list of messages
        :return: The response message for the first message in the *payload*
        :rtype: dict
        :raises TransportError: When the network request fails.
        """
        self._finalize_payload(payload)
        headers = {}
        # process the outgoing payload with the extensions
        await self._process_outgoing_payload(payload, headers)
        # send the payload to the server
        return await self._send_final_payload(payload, headers=headers)

    async def _process_outgoing_payload(self, payload, headers):
        """Process the outgoing *payload* and *headers* with the extensions

        :param list[dict] payload: A list of messages
        :param dict headers: Headers to send
        """
        for extension in self._extensions:
            await extension.outgoing(payload, headers)
        if self._auth:
            await self._auth.outgoing(payload, headers)

    @abstractmethod
    async def _send_final_payload(self, payload, *, headers):
        """Send *payload* to server

        Send the *payload* to the server and return once a
        response message can be provided for the first message in the
        *payload*.
        When reimplementing this method keep in mind that the server will
        likely return additional responses which should be enqueued for
        consumers. To enqueue the received messages :meth:`_consume_payload`
        can be used.

        :param list[dict] payload: A list of messages
        :param dict headers: Headers to send
        :return: The response message for the first message in the *payload*
        :rtype: dict
        :raises TransportError: When the network request fails.
        """

    async def _consume_message(self, response_message):
        """Enqueue the *response_message* for consumers if it's a type of
        message that consumers should receive

        :param response_message: A response message
        """
        if (is_server_error_message(response_message) or
                is_event_message(response_message)):
            await self.incoming_queue.put(response_message)

    def _update_subscriptions(self, response_message):
        """Update the set of subscriptions based on the *response_message*

       :param response_message: A response message
        """
        # if a subscription response is successful, then add the channel
        # to the set of subscriptions, if it fails, then remove it
        if response_message["channel"] == MetaChannel.SUBSCRIBE:
            if (response_message["successful"] and
                    response_message["subscription"]
                    not in self._subscriptions):
                self._subscriptions.add(response_message["subscription"])
            elif (not response_message["successful"] and
                  response_message["subscription"] in self._subscriptions):
                self._subscriptions.remove(response_message["subscription"])

        # if an unsubscribe response is successful then remove the channel
        # from the set of subscriptions
        if response_message["channel"] == MetaChannel.UNSUBSCRIBE:
            if (response_message["successful"] and
                    response_message["subscription"] in self._subscriptions):
                self._subscriptions.remove(response_message["subscription"])

    async def _process_incoming_payload(self, payload, headers=None):
        """Process incoming *payload* and *headers* with the extensions

        :param payload: A list of response messages
        :type payload: list[dict]
        :param headers: Received headers
        :type headers: dict or None
        """
        if self._auth:
            await self._auth.incoming(payload, headers)
        for extension in self._extensions:
            await extension.incoming(payload, headers)

    async def _consume_payload(self, payload, *, headers=None,
                               find_response_for=None):
        """Enqueue event messages for the consumers and update the internal
        state of the transport, based on response messages in the *payload*.

        :param payload: A list of response messages
        :type payload: list[dict]
        :param headers: Received headers
        :type headers: dict or None
        :param dict find_response_for: Find and return the matching \
        response message for the given *find_response_for* message.
        :return: The response message for the *find_response_for* message, \
        otherwise ``None``
        :rtype: dict or None
        """
        # process incoming payload and headers with the extensions
        await self._process_incoming_payload(payload, headers)

        # return None if no response message is found for *find_response_for*
        result = None
        for message in payload:
            # if there is an advice in the message then update the transport's
            # reconnect advice
            if "advice" in message:
                self._reconnect_advice = message["advice"]

            # update subscriptions based on responses
            self._update_subscriptions(message)

            # set the message as the result and continue if it is a matching
            # response
            if (result is None and
                    is_matching_response(message, find_response_for)):
                result = message
                continue

            await self._consume_message(message)
        return result

    def _start_connect_task(self, coro):
        """Wrap the *coro* in a future and schedule it

        The future is stored internally in :obj:`_connect_task`. The future's
        results will be consumed by :obj:`_connect_done`.

        :param coro: Coroutine
        :return: Future
        """
        self._connect_task = asyncio.ensure_future(coro, loop=self._loop)
        self._connect_task.add_done_callback(self._connect_done)
        return self._connect_task

    async def _stop_connect_task(self):
        """Stop the connection task

        If no connect task exists or if it's done it does nothing.
        """
        if self._connect_task and not self._connect_task.done():
            self._connect_task.cancel()
            await asyncio.wait([self._connect_task])

    async def connect(self):
        """Connect to the server

        The transport will try to start and maintain a continuous connection
        with the server, but it'll return with the response of the first
        successful connection as soon as possible.

        :return dict: The response of the first successful connection.
        :raise TransportInvalidOperation: If the transport doesn't has a \
        client id yet, or if it's not in a :obj:`~TransportState.DISCONNECTED`\
        :obj:`state`.
        :raises TransportError: When the network request fails.
        """
        if not self.client_id:
            raise TransportInvalidOperation(
                "Can't connect to the server without a client id. "
                "Do a handshake first.")
        if self.state not in [TransportState.DISCONNECTED,
                              TransportState.SERVER_DISCONNECTED]:
            raise TransportInvalidOperation(
                "Can't connect to a server without disconnecting first.")

        self._state = TransportState.CONNECTING
        return await self._start_connect_task(self._connect())

    async def _connect(self):
        """Connect to the server

        :return: Connect response
        :rtype: dict
        :raises TransportError: When the network request fails.
        """
        message = CONNECT_MESSAGE.copy()
        payload = [message]
        if self._subscribe_on_connect and self.subscriptions:
            for subscription in self.subscriptions:
                extra_message = SUBSCRIBE_MESSAGE.copy()
                extra_message["subscription"] = subscription
                payload.append(extra_message)
        result = await self._send_payload_with_auth(payload)
        self._subscribe_on_connect = not result["successful"]
        return result

    def _connect_done(self, future):
        """Consume the result of the *future* and follow the server's \
        connection advice if the transport is still connected

        :param asyncio.Future future: A :obj:`_connect` or :obj:`handshake` \
        future
        """
        try:
            result = future.result()
            reconnect_timeout = self._reconnect_advice["interval"]
            self._state = TransportState.CONNECTED
        except Exception as error:  # pylint: disable=broad-except
            result = error
            reconnect_timeout = self._reconnect_timeout
            if self.state != TransportState.DISCONNECTING:
                self._state = TransportState.CONNECTING

        LOGGER.debug("Connect task finished with: %r", result)

        if self.state != TransportState.DISCONNECTING:
            self._follow_advice(reconnect_timeout)

    def _follow_advice(self, reconnect_timeout):
        """Follow the server's reconnect advice

        Either a :obj:`_connect` or :obj:`handshake` operation is started
        based on the advice or the method returns without starting any
        operation if no advice is provided.

        :param reconnect_timeout: Initial connection delay to pass to \
        :obj:`_connect` or :obj:`handshake`.
        """
        advice = self._reconnect_advice.get("reconnect")

        # do a handshake operation if advised
        if advice == "handshake":
            handshake_coro = defer(self.handshake,
                                   delay=reconnect_timeout,
                                   loop=self._loop)
            self._start_connect_task(handshake_coro([self.connection_type]))

        # do a connect operation if advised
        elif advice == "retry":
            connect_coro = defer(self._connect,
                                 delay=reconnect_timeout,
                                 loop=self._loop)
            self._start_connect_task(connect_coro())

        # there is not reconnect advice from the server or its value
        # is none
        else:
            LOGGER.warning("No reconnect advice provided, no more operations "
                           "will be scheduled.")
            self._state = TransportState.SERVER_DISCONNECTED

    async def disconnect(self):
        """Disconnect from server

        The disconnect message is only sent to the server if the transport is
        actually connected.
        """
        try:
            should_send_message = self.state == TransportState.CONNECTED

            self._state = TransportState.DISCONNECTING
            await self._stop_connect_task()

            if should_send_message:
                with suppress(TransportError):
                    await self._send_message(DISCONNECT_MESSAGE.copy())
        finally:
            self._state = TransportState.DISCONNECTED

    async def close(self):
        """Close transport and release resources"""
        await self._close_http_session()

    async def subscribe(self, channel):
        """Subscribe to *channel*

        :param str channel: Name of the channel
        :return: Subscribe response
        :rtype: dict
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """
        if self.state not in [TransportState.CONNECTING,
                              TransportState.CONNECTED]:
            raise TransportInvalidOperation(
                "Can't subscribe without being connected to a server.")
        return await self._send_message(SUBSCRIBE_MESSAGE.copy(),
                                        subscription=channel)

    async def unsubscribe(self, channel):
        """Unsubscribe from *channel*

        :param str channel: Name of the channel
        :return: Unsubscribe response
        :rtype: dict
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """
        if self.state not in [TransportState.CONNECTING,
                              TransportState.CONNECTED]:
            raise TransportInvalidOperation(
                "Can't unsubscribe without being connected to a server.")
        return await self._send_message(UNSUBSCRIBE_MESSAGE.copy(),
                                        subscription=channel)

    async def publish(self, channel, data):
        """Publish *data* to the given *channel*

        :param str channel: Name of the channel
        :param dict data: Data to send to the server
        :return: Publish response
        :rtype: dict
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """
        if self.state not in [TransportState.CONNECTING,
                              TransportState.CONNECTED]:
            raise TransportInvalidOperation(
                "Can't publish without being connected to a server.")
        return await self._send_message(PUBLISH_MESSAGE.copy(),
                                        channel=channel,
                                        data=data)
